# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

ALL_PLATFORM = ["torch", "jax", "tvm"]


class MockDeviceMesh:

    def __init(self):
        pass


class TorchMockDeviceMesh(MockDeviceMesh):

    def __init__(self, *arg, debug_only=False):
        super().__init__()
        self.shape = tuple(arg)
        self.debug_only = debug_only

    def size(self, i):
        return self.shape[i]

    def __str__(self) -> str:
        return f"TorchMockDeviceMesh(shape={self.shape})"

    def __repr__(self) -> str:
        return self.__str__()


class JaxDeviceID:

    def __init__(self, *arg):
        self.shape = tuple(arg)


class JaxMockDeviceMesh(MockDeviceMesh):

    def __init__(self, *arg):
        super().__init__()
        self.device_ids = JaxDeviceID(*arg)


def assert_partial_func_equal(func1, func2):
    assert func1.args == func2.args
    assert func1.keywords == func2.keywords
    assert func1.func.__name__ == func2.func.__name__



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from __future__ import annotations
import copy
import os
import logging
from typing import List
from functools import reduce

from easydist.metashard.combination import ReduceOp
from easydist.platform import get_backend
import easydist.config as mdconfig

logger = logging.getLogger(__name__)

DEVICE_MESH_1D = -1


class SPMD:

    REPLICATE = "REPLICATE"
    SHARD = "SHARD"
    PARTIAL = "PARTIAL"

    def __init__(self, state, args=None) -> None:
        self.state = state
        self.args = args

    def is_shard(self) -> bool:
        return self.state == SPMD.SHARD

    def is_replicate(self) -> bool:
        return self.state == SPMD.REPLICATE

    def is_partial(self) -> bool:
        return self.state == SPMD.PARTIAL

    def __eq__(self, other) -> bool:
        if self.state == other.state and self.args == other.args:
            return True
        return False

    def __str__(self) -> str:
        str_ = f"{self.state}"
        if self.args:
            str_ += f"({self.args})"
        return str_

    def __repr__(self) -> str:
        return self.__str__()


class VarSPMDStrategy:

    def __init__(self, *var_spmd_strategy) -> None:
        self.var_spmd_strategy = list(var_spmd_strategy)

    def __getitem__(self, idx) -> SPMD:
        return self.var_spmd_strategy[idx]

    def __add__(self, other):
        return VarSPMDStrategy(*self.var_spmd_strategy, *other.var_spmd_strategy)

    def __eq__(self, other) -> bool:
        if len(self.var_spmd_strategy) == len(other.var_spmd_strategy):
            for i, j in zip(self.var_spmd_strategy, other.var_spmd_strategy):
                if i != j:
                    return False
            return True
        return False

    def __str__(self) -> str:
        return f"VarSPMDStrategy({self.var_spmd_strategy.__str__()})"

    def __repr__(self) -> str:
        return self.__str__()


class VarSPMDStrategyGroup:

    def __init__(self, *var_spmd_strategy_group) -> None:
        self.var_spmd_strategy_group = list(var_spmd_strategy_group)

    def append(self, var_spmd_strategy: VarSPMDStrategy) -> None:
        self.var_spmd_strategy_group.append(var_spmd_strategy)

    def get_var_strtg(self, idx: int) -> VarSPMDStrategy:
        assert idx < len(self.var_spmd_strategy_group)
        return self.var_spmd_strategy_group[idx]

    def __eq__(self, other) -> bool:
        if len(self.var_spmd_strategy_group) == len(other.var_spmd_strategy_group):
            for i, j in zip(self.var_spmd_strategy_group, other.var_spmd_strategy_group):
                if i != j:
                    return False
            return True
        return False

    def __getitem__(self, idx) -> VarSPMDStrategy:
        return self.var_spmd_strategy_group[idx]

    def __str__(self) -> str:
        return f"VarSPMDStrategyGroup({self.var_spmd_strategy_group.__str__()})"

    def __repr__(self) -> str:
        return self.__str__()


class NodeSPMDStrategy:

    def __init__(self, in_strtg_group: VarSPMDStrategyGroup,
                 out_strtg_group: VarSPMDStrategyGroup):
        self.in_strtg_group = in_strtg_group
        self.out_strtg_group = out_strtg_group

    def get_invar_strtg(self, invar_idx: int) -> VarSPMDStrategy:
        return self.in_strtg_group.get_var_strtg(invar_idx)

    def get_outvar_strtg(self, outvar_idx: int) -> VarSPMDStrategy:
        return self.out_strtg_group.get_var_strtg(outvar_idx)

    def __str__(self) -> str:
        return (f"NodeSPMDStrategy(in_strtg_group: {self.in_strtg_group}, "
                f"out_strtg_group: {self.out_strtg_group})")

    def __repr__(self) -> str:
        return self.__str__()


class NodeSPMDStrategyPool:

    def __init__(self):
        self.strategies = []

    def add_strategy(self, strtg: NodeSPMDStrategy):
        self.strategies.append(strtg)

    def find_matched_out(self, out_var_idx: int, expected_strtg: VarSPMDStrategy):
        for strtg_idx, nd_strtg in enumerate(self.strategies):
            if nd_strtg.out_strtg_group.get_var_strtg(out_var_idx) == expected_strtg:
                return strtg_idx

        return -1

    def get_strtg(self, idx: int) -> NodeSPMDStrategy:
        assert idx < len(self.strategies)
        return self.strategies[idx]

    def strtg_num(self) -> int:
        return len(self.strategies)

    def __str__(self) -> str:
        return f"{self.strategies})"

    def __repr__(self) -> str:
        return self.__str__()


def get_sharding_strategy(sharding_anns, shard_dim_id):
    spmd_strategy = VarSPMDStrategyGroup()

    for tensor_ann in sharding_anns.annotation:
        shard_dim_ = None
        for dim_idx, dim_ann in enumerate(tensor_ann):
            if dim_ann.shard_dim_id == shard_dim_id:
                shard_dim_ = dim_idx
        if shard_dim_ is not None:
            spmd_strategy.append(VarSPMDStrategy(SPMD(SPMD.SHARD, {"dim": shard_dim_})))
        else:
            spmd_strategy.append(VarSPMDStrategy(SPMD(SPMD.REPLICATE)))

    return spmd_strategy


def combination_to_sharding_strategy(comm_anns, all_replicate=False):
    # example of comm_anns:
    # functools.partial(<function CombinationFunc.gather at 0x7fab788efd30>, dim=0)

    spmd_strategy = VarSPMDStrategyGroup()

    if not (isinstance(comm_anns, list) or isinstance(comm_anns, tuple)):
        comm_anns = [comm_anns]
    for comm_ann in comm_anns:
        func_name = comm_ann.func.__name__
        if all_replicate or func_name == "identity":
            spmd_strategy.append(VarSPMDStrategy(SPMD(SPMD.REPLICATE)))
        elif func_name == "gather":
            spmd_strategy.append(VarSPMDStrategy(SPMD(SPMD.SHARD, comm_ann.keywords)))
        elif func_name == "reduce":
            spmd_strategy.append(VarSPMDStrategy(SPMD(SPMD.PARTIAL, comm_ann.keywords)))

    return spmd_strategy


_dtype2byte = {
    "float64": 8,
    "float32": 4,
    "float16": 2,
    "bool": 0.125,
    "int32": 4,
    "int64": 8,
    "uint32": 4,
    "uint8": 1,
    "complex64": 16,
}


class MetaVar:

    id_counter: int = 0

    @staticmethod
    def generate_unique_id():
        unique_id = MetaVar.id_counter
        MetaVar.id_counter += 1
        return unique_id

    @staticmethod
    def clear_id_counter():
        MetaVar.id_counter = 0

    def __init__(self, name, shape, dtype) -> None:
        self.unique_id = self.generate_unique_id()
        self.name = name
        self.shape = shape
        self.dtype = dtype
        self.up_node = None
        self.idx_for_up = -1
        self.down_nodes = []
        self.indice_for_down = []

    def get_var_size(self):
        if len(self.shape) == 0:
            return _dtype2byte[self.dtype]
        num_ele = reduce((lambda x, y: x * y), self.shape)
        return _dtype2byte[self.dtype] * num_ele

    def __str__(self, details=False) -> str:
        return self.name.__str__()

    def debug_str(self) -> str:
        return f"{self.name}({self.shape}, {self.dtype})"

    def __repr__(self) -> str:
        return self.__str__()


# (TODO) need to automatic find the heavy ops during MetaSPMD annotation.
_heavy_ops = [
    "convolution",
    "scaled_dot_product",
    "aten.baddbmm",
    "aten.mm",
    "aten.bmm",
    "dot",
]


class MetaNode:

    id_counter: int = 0

    @staticmethod
    def generate_unique_id():
        unique_id = MetaVar.id_counter
        MetaVar.id_counter += 1
        return unique_id

    @staticmethod
    def clear_id_counter():
        MetaVar.id_counter = 0

    def __init__(self, name, op_name, invars: List[MetaVar], outvars: List[MetaVar],
                 sharding_info, is_placeholder=False) -> None:
        self.unique_id = self.generate_unique_id()
        self.cluster_id = -1
        self.name = name
        self.op_name = op_name
        self.invars = invars
        for idx, in_var in enumerate(invars):
            if in_var:
                if isinstance(in_var, MetaVar):  # Lansong(TODO): should remove this line later
                    in_var.down_nodes.append(self)
                    in_var.indice_for_down.append(idx)
        self.compact_out_idx_tbl = []

        self.outvars = outvars
        for idx, out_var in enumerate(outvars):
            if out_var:
                out_var.up_node = self
                out_var.idx_for_up = idx
        self.compact_in_idx_tbl = []

        self.strtg_pool = None
        self.unique_key_ = None

        self.sharding_info = sharding_info
        self.is_placeholder = is_placeholder

    def set_in_var(self, in_var: MetaVar, idx: int):
        assert idx < len(self.compact_in_idx_tbl)
        compact_idx = self.compact_in_idx_tbl[idx]

        assert compact_idx < len(self.invars)
        assert not self.invars[compact_idx]
        self.invars[compact_idx] = in_var
        in_var.down_nodes.append(self)
        in_var.indice_for_down.append(compact_idx)

    def set_out_var(self, out_var: MetaVar, idx: int):
        assert idx < len(self.compact_out_idx_tbl)
        compact_idx = self.compact_out_idx_tbl[idx]

        assert compact_idx < len(self.outvars)
        assert not self.outvars[compact_idx]
        self.outvars[compact_idx] = out_var
        out_var.up_node = self
        out_var.idx_for_up = compact_idx

    def unique_key(self):
        if self.unique_key_ is None:
            self.unique_key_ = str(self.name)
        return self.unique_key_

    def _replicate_strategy(self):
        invars_strategy = VarSPMDStrategyGroup()
        for _ in self.invars:
            invars_strategy.append(VarSPMDStrategy(SPMD(SPMD.REPLICATE)))
        outvars_strategy = VarSPMDStrategyGroup()
        for _ in self.outvars:
            outvars_strategy.append(VarSPMDStrategy(SPMD(SPMD.REPLICATE)))
        return NodeSPMDStrategy(invars_strategy, outvars_strategy)

    def get_strtg(self, idx: int) -> NodeSPMDStrategy:
        assert self.strtg_pool
        return self.strtg_pool.get_strtg(idx)

    def get_strtg_pool(self) -> NodeSPMDStrategyPool:
        if self.strtg_pool is not None:
            return self.strtg_pool

        strategy_list_1d = []

        sharding_anns = self.sharding_info['sharding_ann']
        comm_anns = self.sharding_info['combination_ann']

        # some ops like torch.ops.aten.scalar_tensor.default have no invars
        if len(comm_anns) == 0 and len(self.invars) == 0:
            self.strtg_pool = NodeSPMDStrategyPool()
            return self.strtg_pool

        for comm_ann in comm_anns:
            in_strtg_group = get_sharding_strategy(sharding_anns, comm_ann)
            out_strtg_group = combination_to_sharding_strategy(comm_anns[comm_ann])
            strategy_list_1d.append(NodeSPMDStrategy(in_strtg_group, out_strtg_group))

        if all(op not in self.op_name for op in _heavy_ops) or len(comm_anns) == 0:
            strategy_list_1d.append(self._replicate_strategy())

        # (FIXME) modify stratgy here, need remove
        #     if "batch_norm" in self.name:
        #         dp_strategy = {
        #             "invars_sharding":
        #             get_sharding_strategy(sharding_anns, None),
        #             "outvars_sharding":
        #             combination_to_sharding_strategy(comm_anns[comm_ann], all_replicate=True)
        #         }
        #         dp_strategy["invars_sharding"][0] = SPMD(SPMD.SHARD, {'dim': 0})
        #         dp_strategy["outvars_sharding"][0] = SPMD(SPMD.SHARD, {'dim': 0})
        #         if "backward" in self.name:
        #             dp_strategy["invars_sharding"][1] = SPMD(SPMD.SHARD, {'dim': 0})
        #             dp_strategy["outvars_sharding"][1] = SPMD(SPMD.PARTIAL, {'ops': ReduceOp.SUM})
        #             dp_strategy["outvars_sharding"][2] = SPMD(SPMD.PARTIAL, {'ops': ReduceOp.SUM})
        #         strategy_list = [dp_strategy, replicate_strategy]

        # if "convolution.default" in self.name:
        #     del strategy_list[1]
        # if "convolution_backward.default" in self.name:
        #     del_idx = None
        #     for idx, s in enumerate(strategy_list):
        #         if s["invars_sharding"][-1] == SPMD(SPMD.SHARD, {'dim': 1}):
        #             del_idx = idx
        #             break
        #     if del_idx is not None:
        #         del strategy_list[del_idx]

        self.strtg_pool = NodeSPMDStrategyPool()
        if DEVICE_MESH_1D == -1:
            for idx1, s1 in enumerate(strategy_list_1d):
                for idx2, s2 in enumerate(strategy_list_1d):
                    # [Shard(i), Shard(i)] is not support for pytorch dtensor runtime
                    if get_backend() == "torch":
                        if any(i.state == SPMD.SHARD for i in s1.in_strtg_group) and idx1 == idx2:
                            continue
                    invars_strategy = VarSPMDStrategyGroup()
                    for i, j in zip(s1.in_strtg_group, s2.in_strtg_group):
                        invars_strategy.append(i + j)
                    outvars_strategy = VarSPMDStrategyGroup()
                    for i, j in zip(s1.out_strtg_group, s2.out_strtg_group):
                        outvars_strategy.append(i + j)
                    self.strtg_pool.add_strategy(
                        NodeSPMDStrategy(invars_strategy, outvars_strategy))
        else:
            replicate_ = VarSPMDStrategy(SPMD(SPMD.REPLICATE))
            for s in strategy_list_1d:
                if DEVICE_MESH_1D == 0:
                    invars_strategy = VarSPMDStrategyGroup()
                    for i in s.in_strtg_group:
                        invars_strategy.append(replicate_ + i)
                    outvars_strategy = VarSPMDStrategyGroup()
                    for i in s.out_strtg_group:
                        outvars_strategy.append(replicate_ + i)
                    self.strtg_pool.add_strategy(
                        NodeSPMDStrategy(invars_strategy, outvars_strategy))
                elif DEVICE_MESH_1D == 1:
                    invars_strategy = VarSPMDStrategyGroup()
                    for i in s.in_strtg_group:
                        invars_strategy.append(i + replicate_)
                    outvars_strategy = VarSPMDStrategyGroup()
                    for i in s.out_strtg_group:
                        outvars_strategy.append(i + replicate_)
                    self.strtg_pool.add_strategy(
                        NodeSPMDStrategy(invars_strategy, outvars_strategy))
                else:
                    exit(-1)

        if mdconfig.log_level <= logging.DEBUG:
            print("node: %s" % self.name.__str__())
            print(self.strtg_pool)
        return self.strtg_pool

    def __str__(self) -> str:
        return self.name.__str__()

    def debug_str(self) -> str:
        return_str = f"{self.name} ({self.op_name})\n"
        return_str += "invars: " + ",".join([var.debug_str() for var in self.invars]) + "\n"
        return_str += "outvars: " + ",".join(
            [var.debug_str() if var is not None else "None" for var in self.outvars]) + "\n"
        return_str += f"sharding_info: {self.sharding_info}"
        return return_str

    def __repr__(self) -> str:
        return self.__str__()


class ClusterStrategy:

    def __init__(self):
        self.in_strtg_group = None
        self.out_strtg_group = None
        self.inner_strtg_group = None
        self.node_strategies = {}

    def set_node_strategy(self, nd_id: int, nd_strtg_id: int, nd_strtg: NodeSPMDStrategy):
        assert nd_id not in self.node_strategies

        self.node_strategies[nd_id] = [nd_strtg_id, nd_strtg]

    def __str__(self) -> str:
        return (f"ClusterStrategies(in_strtg_group: {self.in_strtg_group}, "
                f"out_strtg_group: {self.out_strtg_group}, "
                f"inner_strtg_group: {self.inner_strtg_group})")

    def __repr__(self) -> str:
        return self.__str__()


class NodeIOStrategies:

    def __init__(self, node: MetaNode):
        self.node = node
        self.in_strategies = [[] for _ in range(len(node.invars))]
        if self.node.is_placeholder:
            self.in_strategies.append([])
        self.out_strategies = [[] for _ in range(len(node.outvars))]

    def add_in_strategy(self, in_idx: int, var_strtg: VarSPMDStrategy):
        if not self.node.is_placeholder:
            assert in_idx < len(self.node.invars)
        self.in_strategies[in_idx].append(var_strtg)

    def add_out_strategy(self, out_idx: int, var_strtg: VarSPMDStrategy):
        assert out_idx < len(self.node.outvars)
        self.out_strategies[out_idx].append(var_strtg)

    def get_invar_strtg(self, invar_idx: int, strtg_idx: int) -> VarSPMDStrategy:
        return self.in_strategies[invar_idx][strtg_idx]

    def get_outvar_strtg(self, outvar_idx: int, strtg_idx: int) -> VarSPMDStrategy:
        return self.out_strategies[outvar_idx][strtg_idx]

    def get_invar_strtg_list(self, invar_idx: int):  # return list of VarSPMDStrategy
        return self.in_strategies[invar_idx]

    def get_outvar_strtg_list(self, outvar_idx: int):  # return list of VarSPMDStrategy
        return self.out_strategies[outvar_idx]

    def __str__(self) -> str:
        res = str(self.node)
        res += f"\nin strategies: {self.in_strategies}"
        res += f"\noutput strategies: {self.out_strategies}"
        return res

    def __repr__(self) -> str:
        return self.__str__()


class ClusterStrategyPool:

    def __init__(self, cluster: MetaNodeCluster):
        self.cluster = cluster
        self.node_io_strategies = {}
        self.node_strategies = {}
        self.strtg_num = 0

        for nd in cluster.nodes.values():
            self.node_io_strategies[nd.unique_id] = NodeIOStrategies(nd)
            self.node_strategies[nd.unique_id] = []

    def get_invar_strtg(self, nd_id: int, invar_idx: int, strtg_idx: int) -> VarSPMDStrategy:
        assert strtg_idx < self.strtg_num
        nd_io_strtg = self.node_io_strategies[nd_id]
        return nd_io_strtg.get_invar_strtg(invar_idx, strtg_idx)

    def get_outvar_strtg(self, nd_id: int, outvar_idx: int, strtg_idx: int) -> VarSPMDStrategy:
        assert strtg_idx < self.strtg_num
        nd_io_strtg = self.node_io_strategies[nd_id]
        return nd_io_strtg.get_outvar_strtg(outvar_idx, strtg_idx)

    def get_invar_strtg_list(self, nd_id: int, invar_idx: int) -> list[VarSPMDStrategy]:
        nd_io_strtg = self.node_io_strategies[nd_id]
        return nd_io_strtg.get_invar_strtg_list(invar_idx)

    def get_outvar_strtg_list(self, nd_id: int, outvar_idx: int) -> list[VarSPMDStrategy]:
        nd_io_strtg = self.node_io_strategies[nd_id]
        return nd_io_strtg.get_outvar_strtg_list(outvar_idx)

    def get_node_strtg(self, nd_id: int, strtg_idx: int) -> NodeSPMDStrategy:
        assert nd_id in self.node_strategies
        assert strtg_idx < self.strtg_num
        return self.node_strategies[nd_id][strtg_idx]

    def add_strategy(self, strtg: ClusterStrategy):
        self.strtg_num += 1
        for nd_id, nd_strtg_info in strtg.node_strategies.items():
            nd_strtg_id = nd_strtg_info[0]
            nd_strtg = nd_strtg_info[1]
            nd = self.cluster.nodes[nd_id]

            # store node strategies with original format
            self.node_strategies[nd.unique_id].append(nd_strtg)

            # store io strategies
            nd_io_strtg = self.node_io_strategies[nd.unique_id]
            if nd.is_placeholder:
                outvar_strtg = nd_strtg.get_outvar_strtg(0)
                nd_io_strtg.add_in_strategy(0, outvar_strtg)
            else:
                for invar_idx in range(len(nd.invars)):
                    invar_strtg = nd_strtg.get_invar_strtg(invar_idx)
                    nd_io_strtg.add_in_strategy(invar_idx, invar_strtg)

                
            for outvar_idx in range(len(nd.outvars)):
                outvar_strtg = nd_strtg.get_outvar_strtg(outvar_idx)
                nd_io_strtg.add_out_strategy(outvar_idx, outvar_strtg)

    def get_strtg_num(self) -> int:
        return self.strtg_num

    def __str__(self) -> str:
        res = f"node strategies: {self.node_strategies}"
        res += f"\nnode io strategies: {self.node_io_strategies}"
        return res

    def __repr__(self) -> str:
        return self.__str__()


class ClusterArgs:

    def __init__(self) -> None:
        self.descs = []

    def add_arg(self, input_node: MetaNode, idx: int) -> None:
        self.descs.append([input_node, idx])

    def __str__(self) -> str:
        return_str = f"{self.descs.__str__()}"
        return return_str


class MetaNodeCluster:

    def __init__(self, unique_id: int) -> None:
        self.unique_id = unique_id
        self.args = ClusterArgs()
        self.output_node = None
        self.nodes = {}
        self.strategy_pool = None

    def add_node(self, meta_node: MetaNode) -> None:
        assert meta_node.unique_id not in self.nodes
        assert meta_node.unique_id >= 0
        self.nodes[meta_node.unique_id] = meta_node
        meta_node.cluster_id = self.unique_id

    def back_build_strategy(self, nd: MetaNode, nd_strtg_idx: int,
                            cluster_strtg: ClusterStrategy) -> bool:
        succ = True
        for invar_idx, invar in enumerate(nd.invars):
            if not invar:
                continue

            up_node = invar.up_node
            if not up_node:
                continue

            if up_node.unique_id not in self.nodes:
                # up_node is not in the cluster
                continue

            # we should guarantee no multi-output node inside cluster
            idx_for_up = invar.idx_for_up
            assert len(up_node.outvars) == 1
            assert idx_for_up == 0

            nd_strtg = nd.strtg_pool.get_strtg(nd_strtg_idx)
            expected_var_strtg = nd_strtg.get_invar_strtg(invar_idx)

            up_nd_strtg_pool = up_node.get_strtg_pool()
            up_nd_strtg_idx = up_nd_strtg_pool.find_matched_out(idx_for_up, expected_var_strtg)
            if up_nd_strtg_idx >= 0:
                up_nd_strtg = up_nd_strtg_pool.get_strtg(up_nd_strtg_idx)
                cluster_strtg.set_node_strategy(up_node.unique_id, up_nd_strtg_idx, up_nd_strtg)
                if not self.back_build_strategy(up_node, up_nd_strtg_idx, cluster_strtg):
                    succ = False
            else:
                succ = False

        return succ

    def finalize(self) -> None:
        for node in self.nodes.values():
            if node.invars:
                for idx, invar in enumerate(node.invars):
                    if not invar.up_node:
                        self.args.add_arg(node, idx)
                    elif invar.up_node.unique_id not in self.nodes:
                        self.args.add_arg(node, idx)
            else:
                # placeholder or get_attr node
                self.args.add_arg(node, 0)

        all_node_outs = []
        for node in self.nodes.values():
            all_node_outs.extend(node.outvars)

        # collect all variables which are read by outside of cluster
        cluster_out_vars = []
        for var in all_node_outs:
            if var:
                if not var.down_nodes:
                    # output MetaVar of graph has an empty down_nodes list
                    cluster_out_vars.append(var)
                else:
                    for node in var.down_nodes:
                        if node.unique_id not in self.nodes:
                            cluster_out_vars.append(var)

        # get or build strategy info for out node
        out_node = None
        for outvar in cluster_out_vars:
            if not out_node:
                out_node = outvar.up_node
            else:
                # be sure only one output node in a cluster
                assert out_node == outvar.up_node

        self.output_node = out_node

        # build strategy candidates for cluster
        self.strategy_pool = ClusterStrategyPool(self)
        out_strtg_pool = out_node.get_strtg_pool()

        for out_strtg_idx in range(out_strtg_pool.strtg_num()):
            cluster_strtg = ClusterStrategy()
            out_strtg = out_node.get_strtg(out_strtg_idx)
            cluster_strtg.set_node_strategy(out_node.unique_id, out_strtg_idx, out_strtg)

            if self.back_build_strategy(out_node, out_strtg_idx, cluster_strtg):
                assert len(cluster_strtg.node_strategies) == len(self.nodes)
                self.strategy_pool.add_strategy(cluster_strtg)
            else:
                # failed to search sync free cluster strategy
                # Lansong(TODO) we should decompose the cluster into multiple clusters
                logger.debug(
                    "failed to search sync free cluster strategy for node %s, strategy %s",
                    str(out_node), str(out_strtg_pool.get_strtg(out_strtg_idx)))
                continue

    def get_strtg_pool(self) -> ClusterStrategyPool:
        assert self.strategy_pool
        return self.strategy_pool

    def __str__(self) -> str:
        cluster_info = "\ncluster id: " + str(self.unique_id) + "\n  nodes: "
        for node in self.nodes.values():
            cluster_info += str(node) + ", "

        cluster_info += "\n  inputs: " + str(self.args)
        cluster_info += "\n  output: " + str(self.output_node)

        cluster_info += "\nstrategies:\n" + str(self.strategy_pool)
        return cluster_info

    def __repr__(self) -> str:
        return self.__str__()


class MetaGraph:

    def __init__(self, ori_struct) -> None:
        self.ori_struct = ori_struct
        self.input_list = []
        self.op_list = []
        self.output_list = []

        self.node_clusters = []

        self.state_io_map = {}

    def add_input(self, placeholder: MetaNode) -> None:
        self.input_list.append(placeholder)

    def add_node(self, meta_node: MetaNode) -> None:
        self.op_list.append(meta_node)

    def add_output(self, outvars: MetaVar) -> None:
        self.output_list.append(outvars)

    def __str__(self) -> str:
        return_str = f"=====================\n[MetaIR]\n\ninput_list: {self.input_list.__str__()}\n\n"
        for op in self.op_list:
            return_str += f"{op.outvars} <--- [{op.op_name}] --- {op.invars}\n"
        return_str += f"\noutput_list: {self.output_list.__str__()}\n=====================\n"
        return_str += f"\nnode clusters:"
        for cluster in self.node_clusters:
            return_str += str(cluster)
        return_str += "\n=====================\n"
        return return_str

    def __repr__(self) -> str:
        return self.__str__()

    def liveness(self, reserve_input=False):
        liveness_set = set([var.name for var in self.output_list])

        liveness_list = []

        for op in reversed(self.op_list):
            for var in op.invars:
                liveness_set.add(var.name)
            for var in op.outvars:
                if var:  # skip if an output is dangling
                    liveness_set.add(var.name)

            liveness_set_line = copy.deepcopy(liveness_set)
            if reserve_input:
                liveness_set_line.union(set([var.name for var in self.input_list]))

            liveness_list.insert(0, liveness_set_line)

            for var in op.outvars:
                if var:  # skip if an output is dangling
                    liveness_set.remove(var.name)

        return liveness_list

    def build_fine_grain_clusters(self):
        cluster_id = 0
        for node in self.op_list:
            cluster = MetaNodeCluster(unique_id=cluster_id)
            cluster.add_node(node)
            cluster.finalize()
            self.node_clusters.append(cluster)
            cluster_id += 1

    def find_cone_roots(self):
        cone_roots = []
        logger.debug("node num: %s" % len(self.op_list))
        for node in self.op_list:
            down_node_num = 0
            valid_outvars = []
            for out_var in node.outvars:
                if out_var:
                    for down_nd in out_var.down_nodes:
                        if down_nd:
                            valid_outvars.append(out_var)
                            down_node_num += 1

            if down_node_num > 1 or down_node_num == 0:
                cone_roots.append(node)
                continue

            up_node_num = 0
            valid_invars = []
            for in_var in node.invars:
                if in_var and in_var.up_node:
                    valid_invars.append(in_var)
                    up_node_num += 1

            if up_node_num > 1:
                cone_roots.append(node)
                continue
            elif up_node_num == 0:
                continue

            # single up node and single down node
            # Lansong(TODO) we should allow multi-input node as inner node of cone
            assert len(valid_outvars) == 1 and len(valid_invars) == 1
            out_size = valid_outvars[0].get_var_size()
            in_size = valid_invars[0].get_var_size()

            if out_size < in_size:
                cone_roots.append(node)
                continue

        return cone_roots

    def build_cone_cluster(self, nd: MetaNode, root_ids, cluster: MetaNodeCluster):
        cluster.add_node(nd)
        for in_var in nd.invars:
            if in_var and in_var.up_node:
                if in_var.up_node.unique_id not in root_ids:
                    #print("recursive build cone from %s" % str(in_var.up_node))
                    self.build_cone_cluster(in_var.up_node, root_ids, cluster)

    def build_cone_clusters(self):
        cone_roots = self.find_cone_roots()
        logger.debug("root num: %d" % len(cone_roots))
        logger.debug(cone_roots)

        root_ids = set()
        for cone_root in cone_roots:
            root_ids.add(cone_root.unique_id)

        cluster_id = 0
        for cone_root in cone_roots:
            cluster = MetaNodeCluster(unique_id=cluster_id)
            self.build_cone_cluster(cone_root, root_ids, cluster)
            cluster.finalize()
            self.node_clusters.append(cluster)
            cluster_id += 1

    def coarsen(self, coarsen_level: int):
        if coarsen_level == 0:
            self.build_fine_grain_clusters()
        elif coarsen_level == 1:
            self.build_cone_clusters()
        else:
            print("Lansong 2do: support more aggressive coarsening")
            # call cone cluster building instead
            self.build_cone_clusters()

        if mdconfig.log_level <= logging.DEBUG:
            for cluster in self.node_clusters:
                print(str(cluster))

    def dump(self):
        if mdconfig.dump_dir is not None:
            os.makedirs(mdconfig.dump_dir, exist_ok=True)
            filename = os.path.join(mdconfig.dump_dir, "metair.txt")
            with open(filename, "w") as f:
                f.write(self.__str__())
            logger.info(f"MetaIR dump into {filename}")

    def get_input_strategy(self, opt_strategy):

        partial_strategy = {}
        for op in self.op_list:
            if op.is_placeholder:
                if op.unique_key() in opt_strategy:
                    strategy = opt_strategy[op.unique_key()]['strategy'].out_strtg_group[0]
                    partial_strategy[op.outvars[0]] = strategy
                else:
                    logger.warning(f"{op.unique_key()} not found in opt_strategy. (maybe scalar tensor)")

        partial_strategy_list = []

        for var in self.input_list:
            if var in partial_strategy:
                partial_strategy_list.append(partial_strategy[var])
            else:
                partial_strategy_list.append(
                    [SPMD(SPMD.REPLICATE),
                     SPMD(SPMD.REPLICATE)])

        return partial_strategy_list


# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import os
import copy

import torch

import easydist.config as mdconfig


def fix_meta_device(fx_module: torch.fx.GraphModule):

    for node in fx_module.graph.nodes:
        if node.op == 'call_function':
            if "device" in node.kwargs:
                new_kwargs = dict(copy.deepcopy(node.kwargs))
                device = mdconfig.easydist_device
                new_kwargs["device"] = torch.device(device=device)
                assert isinstance(new_kwargs, dict)
                node.kwargs = new_kwargs

    fx_module.recompile()

    return fx_module



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import os
import torch
from torch.fx.node import Node, _get_qualified_name
import torch.distributed.distributed_c10d as c10d

import torch.distributed._tensor as spmd
from torch.distributed._tensor import distribute_tensor
import torch.utils._pytree as pytree

from easydist.metashard.combination import ReduceOp
from easydist.metashard import metair
from easydist.metashard.metair import MetaGraph, MetaNode, MetaVar
from easydist.utils import rsetattr, rgetattr
import easydist.config as mdconfig

from .passes.sharding import get_device_mesh

ABSTRACT_DTYPE = {
    torch.float64: "float64",
    torch.float32: "float32",
    torch.float16: "float16",
    torch.int32: "int32",
    torch.int64: "int64",
    torch.bool: "bool",
    torch.uint8: "uint8",
    torch.complex64: "complex64",
}


def to_torch_spmd(meta_spmd):
    if meta_spmd.state == metair.SPMD.SHARD:
        return spmd.Shard(dim=meta_spmd.args["dim"])
    elif meta_spmd.state == metair.SPMD.PARTIAL:
        mapping_ops = {
            ReduceOp.SUM: c10d.ReduceOp.RedOpType.SUM,
            ReduceOp.MAX: c10d.ReduceOp.RedOpType.MAX,
            ReduceOp.MIN: c10d.ReduceOp.RedOpType.MIN,
        }
        return spmd.placement_types._Partial(reduce_op=mapping_ops[meta_spmd.args["ops"]])
    elif meta_spmd.state == metair.SPMD.REPLICATE:
        return spmd.Replicate()


def materialize(x, device):
    if isinstance(x, torch.Tensor) and x.is_meta:
        if x.dtype == torch.bool:
            return torch.rand(x.size(), dtype=torch.float, device=device) > 0.5
        elif torch.is_floating_point(x):
            return torch.rand(x.size(), dtype=x.dtype, device=device)
        else:
            return torch.randint(high=8, size=x.size(), dtype=x.dtype, device=device)
    return x


def shard_module(model, input_, input_strategy, device="cuda"):
    mesh = get_device_mesh()

    input_strategy = [[to_torch_spmd(i) for i in var_strategy] for var_strategy in input_strategy]

    idx = 0
    for name in dict(model.named_parameters()):

        tensor_data = rgetattr(model, name).data
        tensor_data = materialize(tensor_data, device=device)

        rsetattr(
            model, name,
            torch.nn.parameter.Parameter(distribute_tensor(tensor_data, mesh,
                                                           input_strategy[idx])))

        with torch.no_grad():
            rsetattr(model, name + ".grad", torch.empty_like(rgetattr(model, name).data))
        idx += 1

    for name in dict(model.named_buffers()):

        tensor_data = rgetattr(model, name).data
        tensor_data = materialize(tensor_data, device=device)

        rsetattr(model, name, distribute_tensor(tensor_data, mesh, input_strategy[idx]))
        idx += 1

    shard_input = []
    for tensor in input_:
        tensor = materialize(tensor, device=device)
        shard_input.append(distribute_tensor(tensor, mesh, input_strategy[idx]))
        idx += 1

    return shard_input


def torch2meta_graph(fx_module: torch.fx.GraphModule, state_tensor_num, sharding_info,
                     meta_info) -> MetaGraph:
    meta_graph = MetaGraph(fx_module)
    meta_node_map = {}
    meta_var_map = {}
    MetaNode.clear_id_counter()
    MetaVar.clear_id_counter()
    output_names = []
    # 1. create MetaVar for each output of fx graph node except list or tuple
    #    and create MetaNode for each fx graph node except getitem

    for node in fx_module.graph.nodes:
        if node.op == "call_function":
            # 1.1. create MetaVar
            outvars = []
            compact_out_idx_tbl = []
            compact_out_idx = 0
            # NOTE: list and tuple are followed by getitem
            if isinstance(meta_info[node.name], list) or isinstance(meta_info[node.name], tuple):
                for idx, var_meta in enumerate(meta_info[node.name]):
                    if var_meta is not None and var_meta != {}:
                        compact_out_idx_tbl.append(compact_out_idx)
                        compact_out_idx = compact_out_idx + 1
                    else:
                        compact_out_idx_tbl.append(-1)

                # don't create MetaVar. Will fill it with getitem's MetaVar
                if compact_out_idx > 0:
                    outvars = [None] * compact_out_idx
            else:
                meta_var = MetaVar(name=node.name,
                                   shape=meta_info[node.name]["shape"],
                                   dtype=ABSTRACT_DTYPE[meta_info[node.name]["dtype"]])

                meta_var_map[node.name] = meta_var
                outvars.append(meta_var)

            op_name = _get_qualified_name(node.target)
            if op_name == "_operator.getitem":
                # not create MetaNode
                continue

            # 1.2. create MetaNode
            node_sharding_info = None
            if op_name in sharding_info:
                def _gen_meta(arg):
                    if isinstance(arg, Node):
                        return torch.empty(meta_info[arg.name]["shape"],
                                           dtype=meta_info[arg.name]["dtype"],
                                           device="meta")
                    else:
                        # primitive data type: int, float, etc
                        return arg

                args_meta = pytree.tree_map(_gen_meta, node.args)
                args_meta = str(tuple(args_meta)) + ' | ' + str(node.kwargs)
                if args_meta in sharding_info[op_name]:
                    node_sharding_info = sharding_info[op_name][args_meta]

            node_args_flatten = pytree.tree_flatten(node.args)[0]
            compact_in_idx_tbl = []
            compact_in_idx = 0
            for arg in node_args_flatten:
                if isinstance(arg, Node):
                    compact_in_idx_tbl.append(compact_in_idx)
                    compact_in_idx = compact_in_idx + 1
                else:
                    compact_in_idx_tbl.append(-1)
            invars = [None] * compact_in_idx

            meta_node = MetaNode(name=node.name,
                                 op_name=op_name,
                                 invars=invars,
                                 outvars=outvars,
                                 sharding_info=node_sharding_info)
            meta_node.compact_out_idx_tbl = compact_out_idx_tbl
            meta_node.compact_in_idx_tbl = compact_in_idx_tbl
            meta_node_map[node.name] = meta_node
            meta_graph.add_node(meta_node)
        elif node.op in ["placeholder", "get_attr"]:
            if meta_info[node.name] != {}:
                # 1.1. create MetaVar
                meta_var = MetaVar(name=node.name,
                                   shape=meta_info[node.name]["shape"],
                                   dtype=ABSTRACT_DTYPE[meta_info[node.name]["dtype"]])

                meta_var_map[node.name] = meta_var

                # 1.2. create MetaNode
                node_sharding_info = None
                if node.op in sharding_info:
                    arg_meta_tensor = torch.empty(meta_info[node.name]["shape"],
                                                  dtype=meta_info[node.name]["dtype"],
                                                  device="meta")
                    args_meta = str(arg_meta_tensor)
                    if args_meta in sharding_info[node.op]:
                        node_sharding_info = sharding_info[node.op][args_meta]

                meta_node = MetaNode(name=node.name,
                                     op_name=node.op,
                                     invars=[],
                                     outvars=[meta_var],
                                     sharding_info=node_sharding_info,
                                     is_placeholder=True)
                meta_node.compact_out_idx_tbl = [0]
                meta_node.compact_in_idx_tbl = []
                meta_node_map[node.name] = meta_node
                meta_graph.add_node(meta_node)

                meta_graph.add_input(meta_node)
        elif node.op == "output":
            output_names = [arg.name for arg in node.args[0] if arg is not None]

    # 2. update connection between MetaNode and MetaVar
    for node in fx_module.graph.nodes:
        if node.op == "call_function":
            op_name = _get_qualified_name(node.target)
            if op_name == "_operator.getitem":
                # connect MetaVar of this node to MetaNode of its up node
                arg, idx = node.args
                assert isinstance(arg, Node)
                assert isinstance(idx, int)
                arg_meta_node = meta_node_map[arg.name]  # find meta node of getitem's up node
                meta_var = meta_var_map[node.name]  # find getitem's output var
                arg_meta_node.set_out_var(meta_var, idx)
                continue

            node_args_flatten = pytree.tree_flatten(node.args)[0]
            for idx, arg in enumerate(node_args_flatten):
                if isinstance(arg, Node):
                    meta_node = meta_node_map[node.name]
                    in_var = meta_var_map[arg.name]
                    meta_node.set_in_var(in_var, idx)

    for ouput_n in output_names:
        meta_graph.add_output(meta_var_map[ouput_n])

    # add state io map for input/output sharding mismatch cost
    state_io_map = {}
    for i in range(state_tensor_num):
        state_io_map[meta_graph.input_list[i]] = meta_graph.output_list[i]
    meta_graph.state_io_map = state_io_map

    meta_graph.coarsen(coarsen_level=mdconfig.coarsen_level)

    return meta_graph


def get_torch_sharding_strategy(fx_module: torch.fx.GraphModule, opt_strategy):
    sharding_strategy = {}

    for node in fx_module.graph.nodes:
        if node.op == "call_function":
            op_name = _get_qualified_name(node.target)
            if op_name != "_operator.getitem":
                unique_key = node.name
                if unique_key in opt_strategy:
                    sharding_strategy[node.name] = [[
                        to_torch_spmd(ii) for ii in i
                    ] for i in opt_strategy[unique_key]['strategy'].in_strtg_group]
                else:
                    unique_key = f"{op_name}_[{node.name}]"
                    if unique_key in opt_strategy:
                        sharding_strategy[node.name] = [[
                            to_torch_spmd(ii) for ii in i
                        ] for i in opt_strategy[unique_key]['strategy'].in_strtg_group]

    return sharding_strategy



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import torch
from torch.fx.node import _get_qualified_name


def fix_addmm_bias(fx_module: torch.fx.GraphModule):

    for node in fx_module.graph.nodes:
        if node.op == 'call_function':
            if "torch.ops.aten.addmm.default" in _get_qualified_name(node.target):
                node.target = torch.ops.aten.mm.default
                bias = node.args[0]
                node.args = (node.args[1], node.args[2])

                with fx_module.graph.inserting_after(node):
                    add_bias_node = fx_module.graph.call_function(torch.ops.aten.add.Tensor,
                                                                  args=(node, bias))

                    node.replace_all_uses_with(add_bias_node)

                    add_bias_node.update_arg(0, node)

    fx_module.recompile()

    return fx_module


def fix_convoluation_bias(fx_module: torch.fx.GraphModule):

    for node in fx_module.graph.nodes:
        if node.op == 'call_function':
            if "torch.ops.aten.convolution.default" in _get_qualified_name(node.target):
                if node.args[2] is not None:
                    node.target = torch.ops.aten.convolution.default
                    bias = node.args[2]
                    node.args = (node.args[0], node.args[1], None, *node.args[3:])

                    with fx_module.graph.inserting_after(node):
                        bias_new = fx_module.graph.call_function(torch.ops.aten.view.default,
                                                                 args=(bias, [1, -1, 1, 1]))

                    with fx_module.graph.inserting_after(bias_new):
                        add_bias_node = fx_module.graph.call_function(torch.ops.aten.add.Tensor,
                                                                      args=(node, bias_new))

                        node.replace_all_uses_with(add_bias_node)

                        add_bias_node.update_arg(0, node)

    fx_module.recompile()

    return fx_module



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from functools import partial

import torch
import torch.utils._pytree as pytree
from torch._subclasses import FakeTensor
from torch.utils._mode_utils import no_dispatch


def fake_to_real(e):
    with no_dispatch():
        if isinstance(e, FakeTensor):
            out = torch.zeros_like(e, device=e.fake_device)
            return out
        return e


@torch.no_grad()
def materialize_zero(tensor, materialization_device):
    tensor = fake_to_real(tensor)
    return torch.zeros_like(tensor, device=materialization_device)


@torch.no_grad()
def materialize_random(tensor, materialization_device):
    if tensor.dtype == torch.bool:
        return torch.rand(tensor.size(), dtype=torch.float, device=materialization_device) > 0.5
    elif torch.is_floating_point(tensor):
        return torch.rand(tensor.size(), dtype=tensor.dtype, device=materialization_device)
    else:
        return torch.randint(high=8,
                             size=tensor.size(),
                             dtype=tensor.dtype,
                             device=materialization_device)


@torch.no_grad()
def materialize_module_from_cpu(tensor, param_buf_key, cpu_model: torch.nn.Module,
                                materialization_device):
    cpu_params_buffers = {
        **dict(cpu_model.named_parameters()),
        **dict(cpu_model.named_buffers()),
    }
    if param_buf_key in cpu_params_buffers:
        return cpu_params_buffers[param_buf_key].to(device=materialization_device)

    return materialize_zero(tensor, materialization_device)


@torch.no_grad()
def materialize_module_from_set_parameter(tensor, param_buf_key, module: torch.nn.Module,
                                          materialization_device):
    tensor = fake_to_real(tensor)

    submodule_name_list = param_buf_key.split(".")

    if "reset_parameters" in module.__dir__():
        if param_buf_key in dict(module.named_parameters()):
            ori_parameters = module._parameters
            module._parameters = pytree.tree_map(fake_to_real, module._parameters)
            module.to_empty(device=materialization_device)
            module.reset_parameters()
            para = dict(module.named_parameters()).get(param_buf_key, None)
            module._parameters = ori_parameters
            return para

    if "reset_buffers" in module.__dir__():
        if param_buf_key in dict(module.named_buffers()):
            ori_buffers = module._buffers
            module._buffers = pytree.tree_map(fake_to_real, module._buffers)
            module.reset_buffers()
            buf = dict(module.named_buffers()).get(param_buf_key, None)
            module._buffers = ori_buffers
            return buf

    if len(submodule_name_list) == 1:
        return materialize_zero(tensor, materialization_device)

    submodule_name, sub_param_buf_key = submodule_name_list[0], ".".join(submodule_name_list[1:])

    if submodule_name in module._modules:
        return materialize_module_from_set_parameter(tensor, sub_param_buf_key,
                                                     module._modules[submodule_name],
                                                     materialization_device)

    return materialize_zero(tensor, materialization_device)


class InitHelper:

    def __init__(self) -> None:
        pass

    def get_materialize_fn(self):
        pass


class SetParaInitHelper:

    def __init__(self, module=None) -> None:
        self.module = module

    def get_materialize_fn(self):
        return partial(materialize_module_from_set_parameter, module=self.module)


class CpuModuleInitHelper:

    def __init__(self, cpu_module) -> None:
        self.cpu_module = cpu_module

    def get_materialize_fn(self):
        return partial(materialize_module_from_cpu, cpu_model=self.cpu_module)


def init_contiguous_buf(params, params_strategy, device_mesh):

    params = list(params.values())
    dtype = params[0].dtype
    device = params[0].device
    if not all(p.dtype == dtype for p in params):
        raise ValueError("All parameters must be of the same dtype.")
    if not all(p.device == device for p in params):
        raise ValueError("All parameters must be on the same device.")
    contiguous_buf_size = 0
    assert len(params_strategy) == len(params), "mismatch length of params_strategy and params"
    for p_s, p in zip(params_strategy, params):
        tensor_size = p.numel()
        for mesh_dim_idx, placement in enumerate(p_s):
            if placement.is_shard():
                tensor_size = (tensor_size // device_mesh.mesh.shape[mesh_dim_idx]) + 1

        contiguous_buf_size += tensor_size

    return torch.empty(contiguous_buf_size, dtype=dtype, device=device)



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import functools
import operator

from easydist.metashard.annotation import ShardAnnotation, ShardDim
from easydist.metashard.combination import CombinationFunc

EXTEND_VIEW = False


def get_next_non_one(shape_, idx_):
    if idx_ < len(shape_):
        while shape_[idx_] == 1:
            idx_ += 1
            if idx_ >= len(shape_):
                break
    return idx_


def view_propagation(input_shape, output_shape, world_size=1):

    if -1 in output_shape:
        numel = functools.reduce(operator.mul, input_shape)
        dim_size = -1 * numel // functools.reduce(operator.mul, output_shape)
        output_shape[output_shape.index(-1)] = dim_size

    sharding_ann = ShardAnnotation([[ShardDim(0)] * len(input_shape)])
    combination_ann = {}

    input_idx = get_next_non_one(input_shape, 0)
    output_idx = get_next_non_one(output_shape, 0)

    shard_dim = 1

    while input_idx < len(input_shape):
        if input_shape[input_idx] == output_shape[output_idx]:
            # [**, A, **] -> [**, A, **]
            if input_shape[input_idx] >= world_size:
                sharding_ann[0][input_idx] = ShardDim(shard_dim)
                combination_ann[shard_dim] = functools.partial(CombinationFunc.gather,
                                                               dim=output_idx)
                shard_dim += 1
            input_idx = get_next_non_one(input_shape, input_idx + 1)
            output_idx = get_next_non_one(output_shape, output_idx + 1)
        elif input_shape[input_idx] > output_shape[output_idx]:
            # [**, A, **] -> [**, a1, a2, **]
            leftmost_idx = output_idx
            accum_shape_ = output_shape[output_idx]
            for o_idx in range(output_idx + 1, len(output_shape)):
                accum_shape_ *= output_shape[o_idx]
                if accum_shape_ == input_shape[input_idx]:
                    if output_shape[leftmost_idx] >= world_size:
                        sharding_ann[0][input_idx] = ShardDim(shard_dim)
                        combination_ann[shard_dim] = functools.partial(CombinationFunc.gather,
                                                                       dim=leftmost_idx)
                        shard_dim += 1
                    output_idx = get_next_non_one(output_shape, o_idx + 1)
                    input_idx = get_next_non_one(input_shape, input_idx + 1)
                    break
        else:
            # [**, a1, a2, **] -> [**, A, **]
            leftmost_idx = input_idx
            accum_shape_ = input_shape[input_idx]
            for i_idx in range(input_idx + 1, len(input_shape)):
                accum_shape_ *= input_shape[i_idx]
                if accum_shape_ == output_shape[output_idx]:
                    if EXTEND_VIEW:
                        chunk_size_ = 1
                        for sub_idx in range(input_idx, i_idx + 1):
                            sharding_ann[0][sub_idx] = ShardDim(shard_dim)
                            combination_ann[shard_dim] = functools.partial(CombinationFunc.gather,
                                                                           dim=output_idx,
                                                                           chunk=chunk_size_)
                            chunk_size_ *= input_shape[sub_idx]
                            shard_dim += 1
                    else:
                        if input_shape[input_idx] >= world_size:
                            sharding_ann[0][input_idx] = ShardDim(shard_dim)
                            combination_ann[shard_dim] = functools.partial(CombinationFunc.gather,
                                                                           dim=output_idx)
                            shard_dim += 1

                    output_idx = get_next_non_one(output_shape, output_idx + 1)
                    input_idx = get_next_non_one(input_shape, i_idx + 1)
                    break

    return {'sharding_ann': sharding_ann, 'combination_ann': combination_ann}


def view_propagation_preset(input_shape, output_shape, preset_anno):
    accum_size = 1
    for idx, ann in enumerate(preset_anno[0]):
        if ann.shard_dim_id != 0:
            break
        accum_size *= input_shape[idx]

    chunk = preset_anno[0][idx].chunk

    out_accum_size = 1
    out_idx = 0
    while out_accum_size < accum_size:
        out_accum_size *= output_shape[out_idx]
        out_idx += 1

    if out_accum_size == accum_size:
        accum_chunk = 1
        if chunk == accum_chunk:
            return functools.partial(CombinationFunc.gather, dim=out_idx)
        for o_idx in range(out_idx, len(output_shape)):
            if chunk == accum_chunk:
                return functools.partial(CombinationFunc.gather, dim=o_idx)
            accum_chunk *= output_shape[o_idx]



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import time
import logging

import rich
import torch
from torch._functorch.aot_autograd import aot_function, make_boxed_compiler

from easydist.metashard import metair
from easydist.autoflow.solver import AutoFlowSolver
import easydist.config as mdconfig
from .sharding_interpreter import EDTorchShardingAnn
from .utils import _sharding_ann_env
from .bridge import torch2meta_graph, get_torch_sharding_strategy
from .passes import eliminate_detach, fix_addmm_bias, fix_convoluation_bias, fix_embedding, sharding_transform
from .passes.sharding import get_device_mesh

logger = logging.getLogger(__name__)

ENABLE_TRANSFORM = False
BW_CONSTRAINTS = None
INPUT_STRATEGY = None


def get_input_strategy():
    global INPUT_STRATEGY
    return INPUT_STRATEGY


def enable_transform():
    global ENABLE_TRANSFORM
    ENABLE_TRANSFORM = True


def _get_output_strategy(opt_strategy, meta_graph, input_strategy):
    partial_strategy = {}
    for op in meta_graph.op_list:
        op_key = op.unique_key()
        if op_key in opt_strategy:
            for idx, var in enumerate(op.outvars):
                if var in meta_graph.output_list:
                    strategy = opt_strategy[op_key]['strategy'].out_strtg_group.get_var_strtg(idx)
                    partial_strategy[var] = strategy

    for var in meta_graph.input_list:
        if var in meta_graph.output_list:
            if var in input_strategy:
                partial_strategy[var] = input_strategy[var]

    return partial_strategy


def _get_input_strategy(opt_strategy, meta_graph):
    partial_strategy = {}
    for op in reversed(meta_graph.op_list):
        op_key = op.unique_key()
        if op_key in opt_strategy:
            for idx, var in enumerate(op.invars):
                if var in meta_graph.input_list:
                    strategy = opt_strategy[op_key]['strategy'].in_strtg_group.get_var_strtg(idx)
                    partial_strategy[var] = strategy

    partial_strategy_list = []

    for var in meta_graph.input_list:
        if var in partial_strategy:
            partial_strategy_list.append(partial_strategy[var])
        else:
            partial_strategy_list.append(
                [metair.SPMD(metair.SPMD.REPLICATE),
                 metair.SPMD(metair.SPMD.REPLICATE)])

    return partial_strategy, partial_strategy_list


@make_boxed_compiler
def easydist_shard(fx_module: torch.fx.GraphModule, inps):

    global BW_CONSTRAINTS, INPUT_STRATEGY

    fx_module = fix_embedding(fx_module)
    fx_module = fix_addmm_bias(fx_module)
    fx_module = fix_convoluation_bias(fx_module)
    fx_module = eliminate_detach(fx_module)
    fx_module.recompile()
    if mdconfig.log_level <= logging.DEBUG:
        print(fx_module.graph)

    with _sharding_ann_env():
        start_t = time.perf_counter()
        sharding_interpreter = EDTorchShardingAnn(fx_module)
        sharding_info, fwd_shape_info = sharding_interpreter.run(*inps)
        logger.info(f"[EDTorchShardingAnn.time]:\t {time.perf_counter() - start_t} s.")
        if mdconfig.log_level <= logging.DEBUG:
            rich.print("sharding_info:\n", sharding_info)
            rich.print("fwd_shape_info:\n", fwd_shape_info)

    meta_graph = torch2meta_graph(fx_module, sharding_info, fwd_shape_info)

    if mdconfig.log_level <= logging.DEBUG:
        rich.print(meta_graph)

    device_mesh = get_device_mesh()
    device_mesh_shape = (device_mesh.size(0), device_mesh.size(1))

    solver = AutoFlowSolver(device_mesh_shape, BW_CONSTRAINTS)
    solver.add_graph(meta_graph)
    start_t = time.perf_counter()
    count_invars = BW_CONSTRAINTS is None
    opt_strategy = solver.ilp_optimize(count_invars)
    logger.info(f"[AutoFlowSolver.time]:\t {time.perf_counter() - start_t} s.")
    # start_t = time.perf_counter()
    # beam_search_strategy = solver.beam_search()
    # logger.info(f"[AutoFlowSolver.beam_search]: {time.perf_counter() - start_t} s.")

    if mdconfig.log_level <= logging.DEBUG:
        rich.print(opt_strategy)

    if BW_CONSTRAINTS is None:
        strategy_map, INPUT_STRATEGY = _get_input_strategy(opt_strategy, meta_graph)
        BW_CONSTRAINTS = _get_output_strategy(opt_strategy, meta_graph, strategy_map)

    if ENABLE_TRANSFORM:
        sharding_strategy = get_torch_sharding_strategy(fx_module, opt_strategy)

        if mdconfig.log_level <= logging.DEBUG:
            print(sharding_strategy)

        fx_module = sharding_transform(fx_module, sharding_strategy)
        fx_module = fix_embedding(fx_module, recover=True)

        if mdconfig.log_level <= logging.DEBUG:
            print(fx_module.graph)

    return fx_module


def md_aot_module(mod: torch.nn.Module, *args, **kwargs) -> torch.nn.Module:
    """
    Traces the forward and backward graph of :attr:`mod` using torch dispatch
    tracing mechanism. It is wrapper function, that underneath uses
    :func:`aot_function` to perform tracing and compilation.

    :func:`aot_module` lifts the parameters and buffers of ``nn.Module`` as inputs
    to a new callable which is then compiled through :func:`aot_function`.

    .. warning::
        This API is experimental and likely to change.

    Args:
        mod (Callable): A ``nn.Module`` module.
        args : args to be passed to :func:`aot_function`
        kwargs : kwargs to be passed to :func:`aot_function`

    Returns:
        Returns a ``nn.Module`` that retains the eager behavior of the original
        :attr:`mod`, but with forward and backward graph compiled.

    """
    # See Note: [Fake Modules and AOTAutograd]
    torch._dynamo.utils.assert_no_fake_params_or_buffers(mod)

    def functional_call(named_params, named_buffers, *args, **kwargs):
        params_and_buffers = {**named_params, **named_buffers}
        return torch.func.functional_call(mod, params_and_buffers, args, kwargs)

    named_params = dict(mod.named_parameters(remove_duplicate=False))
    named_buffers = dict(mod.named_buffers(remove_duplicate=False))
    num_params_buffers = len(named_params) + len(named_buffers)
    compiled_f = aot_function(functional_call,
                              num_params_buffers=num_params_buffers,
                              *args,
                              **kwargs)

    class AOTModule(torch.nn.Module):

        def __init__(self):
            super().__init__()
            self.orig_module = mod

        def forward(self, *args, **kwargs):
            named_params = dict(mod.named_parameters(remove_duplicate=False))
            named_buffers = dict(mod.named_buffers(remove_duplicate=False))
            return compiled_f(
                named_params,
                named_buffers,
                *args,
                **kwargs,
            )

    return AOTModule()


def compile(mod: torch.nn.Module):
    return md_aot_module(mod, fw_compiler=easydist_shard, bw_compiler=easydist_shard)



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import copy
import logging
import os
from typing import Any, Dict, Iterator, Tuple

import torch
import torch.utils._pytree as pytree
from torch.fx.graph_module import GraphModule
from torch.fx.interpreter import Interpreter
from torch.fx.node import Argument, Target, _get_qualified_name
from torch._subclasses.fake_tensor import FakeTensorMode

from easydist.metashard import MetaOp
from easydist.torch.utils import to_meta
import easydist.config as mdconfig
from .preset_propagation import preset_meta_spmd
from .device_mesh import device_mesh_world_size

logger = logging.getLogger(__name__)


def get_shape_info(node_output):
    if isinstance(node_output, torch.Tensor) or isinstance(node_output,
                                                           torch.nn.parameter.Parameter):
        return {"shape": node_output.shape, "dtype": node_output.dtype}
    else:
        return {}


def meta_exec(self, flat_meta_input=None, enable_fallback=True):

    fake_tensor_mode = FakeTensorMode()

    def wrap_fake(x):
        if isinstance(x, torch.Tensor):
            return fake_tensor_mode.from_tensor(x)
        return x

    if flat_meta_input is None:
        flat_meta_input = self.flat_input_args
    try:
        fake_flat_args = pytree.tree_map(wrap_fake, flat_meta_input)
        fake_args, fake_kwargs = pytree.tree_unflatten(fake_flat_args, self.input_pytree_spec)
        with fake_tensor_mode:
            out = self.func(*fake_args, **fake_kwargs)
    except:
        if enable_fallback:
            out = self.exec()
        else:
            raise RuntimeError(f"failed meta_exec with MetaOp {self}")
    return out


# register the meta_exec function in MetaOp
MetaOp.meta_exec = meta_exec


def to_real(tensor, size=None):

    device = mdconfig.easydist_device

    if size is None:
        size = tensor.size()

    if isinstance(tensor, torch.Tensor) and tensor.is_meta:
        if tensor.dtype == torch.bool:
            return torch.rand(size, dtype=torch.float, device=device) > 0.5
        elif torch.is_floating_point(tensor):
            return torch.rand(size, dtype=tensor.dtype, device=device)
        else:
            return torch.randint(high=1, size=size, dtype=tensor.dtype, device=device)
    return tensor


class EDTorchShardingAnn(Interpreter):

    def __init__(self, module: GraphModule, use_cache=True, garbage_collect_values: bool = True):
        super().__init__(module, garbage_collect_values)
        self.use_cache = use_cache
        self.shape_info = {}
        self.sharding_info = {}

        self.pass_by_ops = ["_operator.getitem"]

    def run(self, *args) -> Any:
        """
        Run `module` via interpretation and return the result.
        Args:
            *args: The arguments to the Module to run, in positional order
            initial_env (Optional[Dict[Node, Any]]): An optional starting environment for execution.
                This is a dict mapping `Node` to any value. This can be used, for example, to
                pre-populate results for certain `Nodes` so as to do only partial evaluation within
                the interpreter.
        Returns:
            Any: The value returned from executing the Module
        """
        self.env = {}

        args = pytree.tree_map(to_meta, args)

        # Positional function args are consumed left-to-right by
        # `placeholder` nodes. Use an iterator to keep track of
        # position and extract those values.
        self.args_iter: Iterator[Any] = iter(args)

        for node in self.module.graph.nodes:

            if node in self.env:
                # Short circuit if we have this value. This could
                # be used, for example, for partial evaluation
                # where the caller has pre-populated `env` with
                # values for a subset of the program.
                continue

            node_output = self.run_node(node)
            self.env[node] = pytree.tree_map(to_meta, node_output)
            self.shape_info[node.name] = pytree.tree_map(get_shape_info, self.env[node])

            if self.garbage_collect_values:
                for to_delete in self.user_to_last_uses.get(node, []):
                    del self.env[to_delete]

            if node.op == 'output':
                # output_val = self.env[node]
                self.env = {}
                return self.sharding_info, self.shape_info

    def placeholder(self, target : 'Target', args : Tuple[Argument, ...], kwargs : Dict[str, Any]) -> Any:
        """
        Execute a ``placeholder`` node. Note that this is stateful:
        ``Interpreter`` maintains an internal iterator over
        arguments passed to ``run`` and this method returns
        next() on that iterator.

        Args:
            target (Target): The call target for this node. See
                `Node <https://pytorch.org/docs/master/fx.html#torch.fx.Node>`__ for
                details on semantics
            args (Tuple): Tuple of positional args for this invocation
            kwargs (Dict): Dict of keyword arguments for this invocation

        Returns:
            Any: The argument value that was retrieved.
        """
        assert isinstance(target, str)

        if "placeholder" not in self.sharding_info:
            self.sharding_info["placeholder"] = {}

        if target.startswith('*'):
            # For a starred parameter e.g. `*args`, retrieve all
            # remaining values from the args list.
            return list(self.args_iter)
        else:
            try:
                output = next(self.args_iter)
                if isinstance(output, torch.Tensor):
                    out_meta_str = str(output)
                    if out_meta_str not in self.sharding_info["placeholder"]:
                        sharding_ann, combination_ann = preset_meta_spmd("placeholder", (output, kwargs))
                        if sharding_ann or combination_ann:
                            self.sharding_info["placeholder"][out_meta_str] = {
                                "sharding_ann": sharding_ann,
                                "combination_ann": combination_ann,
                            }
                return output
            except StopIteration as si:
                if len(args) > 0:
                    return args[0]
                else:
                    raise RuntimeError(f'Expected positional argument for parameter {target}, but one was not passed in!') from si

    def get_attr(self, target : 'Target', args : Tuple[Argument, ...], kwargs : Dict[str, Any]) -> Any:
        """
        Execute a ``get_attr`` node. Will retrieve an attribute
        value from the ``Module`` hierarchy of ``self.module``.

        Args:
            target (Target): The call target for this node. See
                `Node <https://pytorch.org/docs/master/fx.html#torch.fx.Node>`__ for
                details on semantics
            args (Tuple): Tuple of positional args for this invocation
            kwargs (Dict): Dict of keyword arguments for this invocation

        Return:
            Any: The value of the attribute that was retrieved
        """
        assert isinstance(target, str)
        output = self.fetch_attr(target)

        if "get_attr" not in self.sharding_info:
            self.sharding_info["get_attr"] = {}

        if isinstance(output, torch.Tensor):
            out_meta_str = str(output)
            if out_meta_str not in self.sharding_info["get_attr"]:
                sharding_ann, combination_ann = preset_meta_spmd("get_attr", (output, kwargs))
                if sharding_ann or combination_ann:
                    self.sharding_info["get_attr"][out_meta_str] = {
                        "sharding_ann": sharding_ann,
                        "combination_ann": combination_ann,
                    }
        return output

    def call_function(self, target: 'Target', args: Tuple[Argument, ...],
                      kwargs: Dict[str, Any]) -> Any:
        """
        Execute a ``call_function`` node and return the result.
        Args:
            target (Target): The call target for this node. See
                `Node <https://pytorch.org/docs/master/fx.html#torch.fx.Node>`__ for
                details on semantics
            args (Tuple): Tuple of positional args for this invocation
            kwargs (Dict): Dict of keyword arguments for this invocation
        Return
            Any: The value returned by the function invocation
        """
        assert not isinstance(target, str)

        ops_name = _get_qualified_name(target)

        if ops_name in self.pass_by_ops:
            real_out = target(*args, **kwargs)
            return real_out

        args_meta, kwargs_meta = args, kwargs

        op_perf_key = {
            "ops_name": ops_name,
            "args_meta": str(args_meta) + ' | ' + str(kwargs_meta)
        }
        logger.debug(op_perf_key)

        def get_hint_size(args_list, mock_dim_value=1024):
            size_dict = {}
            element_number = 0
            max_dim_value = 0
            for arg_idx in range(len(args_list)):
                if isinstance(args_list[arg_idx], torch.Tensor):
                    tensor_size = args_list[arg_idx].size()
                    size_dict[arg_idx] = tensor_size
                    element_number += tensor_size.numel()
                    if len(args_list[arg_idx].size()) > 0:
                        max_dim_value = max(max_dim_value, max(tensor_size))

            # if element number of all tensor larger then 1024 ** 3,
            hint_size = size_dict
            if mdconfig.use_hint and element_number >= 1024**3:
                hint_size = {}
                for arg_idx in size_dict:
                    hint_size[arg_idx] = torch.Size([
                        mock_dim_value if dim == max_dim_value else dim
                        for dim in size_dict[arg_idx]
                    ])

                logger.info(f"use mock shape for {op_perf_key}, {size_dict} -> {hint_size}")

            return size_dict, hint_size

        # (CAUTION !!) use mock (smaller size) args/kwargs here for OOM issue of giant Op.
        # materialize the input of ops, execute the function and return the result
        def materialize_args_kwargs(args, kwargs):
            flat_args, args_specs = pytree.tree_flatten((args, kwargs))

            ori_size, hint_size = get_hint_size(flat_args)

            def materialize_with_size_dict(size_dict):
                for arg_idx in range(len(flat_args)):
                    if isinstance(flat_args[arg_idx], torch.Tensor):
                        real_tensor = to_real(flat_args[arg_idx], size_dict[arg_idx])

                        # clip for embedding op prevent indexerror in CUDA kernel.
                        if ops_name in [
                                "torch.ops.aten.embedding.default",
                                "easydist.torch.passes.fix_embedding.md_embedding"
                        ]:
                            real_tensor.clip_(max=args[0].shape[0])

                        flat_args[arg_idx] = real_tensor

            materialize_with_size_dict(hint_size)

            # try to run the metaop on the mock size (hint_size)
            # if failed with materialize with the original size
            try:
                meta_op = MetaOp(func=target, input_args=(args, kwargs), name=ops_name)
                meta_op.meta_exec(flat_meta_input=flat_args)
            except:
                materialize_with_size_dict(ori_size)

            return pytree.tree_unflatten(flat_args, args_specs)

        args, kwargs = materialize_args_kwargs(args_meta, kwargs_meta)
        meta_op = MetaOp(func=target,
                         input_args=(args, kwargs),
                         shard_size=device_mesh_world_size(),
                         name=ops_name)
        # user fake tensor here, maybe use shape/dtype info from `make_fx`
        meta_out = meta_op.meta_exec(flat_meta_input=pytree.tree_flatten((args_meta,
                                                                          kwargs_meta))[0])

        # sharding discovery
        if op_perf_key["ops_name"] not in self.sharding_info:
            self.sharding_info[op_perf_key["ops_name"]] = {}

        if not self.use_cache or op_perf_key["args_meta"] not in self.sharding_info[
                op_perf_key["ops_name"]]:
            prompt_annotation = None
            if self.use_cache and len(self.sharding_info[op_perf_key["ops_name"]]) >= 1:
                prompt_annotation = list(
                    self.sharding_info[op_perf_key["ops_name"]].values())[0]["sharding_ann"]

            sharding_ann, combination_ann = preset_meta_spmd(meta_op, (args_meta, kwargs_meta))
            if sharding_ann is None and combination_ann is None:
                sharding_ann, combination_ann = meta_op.sharding_discovery(
                    prompt_annotation=copy.deepcopy(prompt_annotation))

            self.sharding_info[op_perf_key["ops_name"]][op_perf_key["args_meta"]] = {
                "sharding_ann": sharding_ann,
                "combination_ann": combination_ann,
            }

        return pytree.tree_map(to_meta, meta_out)



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import hashlib
from contextlib import contextmanager
from copy import copy
from typing import Any, Dict

import torch
from torch._subclasses.fake_tensor import FakeTensor
from torch.utils._mode_utils import no_dispatch
import torch.utils._pytree as pytree


def to_meta(node_output):
    if isinstance(node_output, FakeTensor):
        with no_dispatch():
            return torch.zeros_like(node_output, device="meta")
    if type(node_output) is torch.Tensor:
        return node_output.detach().to(device="meta").contiguous()
    elif type(node_output) is torch.nn.parameter.Parameter:
        return node_output.data.detach().to(device="meta").contiguous()
    else:
        return node_output


@contextmanager
def _sharding_ann_env():

    ori_cudnn_enabled = torch.backends.cudnn.enabled
    ori_matmul_allow_tf32 = torch.backends.cuda.matmul.allow_tf32
    ori_cudnn_allow_tf32 = torch.backends.cudnn.allow_tf32

    # (TODO) device failed when convolution_backward in sharding_discovery
    torch.backends.cudnn.enabled = False

    torch.backends.cuda.matmul.allow_tf32 = True
    torch.backends.cudnn.allow_tf32 = True

    try:
        yield
    finally:
        torch.backends.cudnn.enabled = ori_cudnn_enabled
        torch.backends.cuda.matmul.allow_tf32 = ori_matmul_allow_tf32
        torch.backends.cudnn.allow_tf32 = ori_cudnn_allow_tf32


@contextmanager
def _rematerialize_optimizer(
    opt: torch.optim.Optimizer,
    named_states: Dict[str, Any],
    params: Dict[str, torch.nn.Parameter],
):
    assert opt is not None

    # update opt.state with proxy tensors
    orig_states: Dict[str, Any] = copy(opt.state)
    for n in named_states:
        # opt.state's key type is string, but optimizer uses Parameter as keys
        opt.state[params[n]] = named_states[n]  # type: ignore[index]

    # FIXME: support multiple parameter groups
    param_group = opt.param_groups[0]
    orig_params = param_group["params"]
    param_group["params"] = params.values()

    try:
        yield
    finally:
        param_group["params"] = orig_params
        opt.state.update(orig_states)


@contextmanager
def _enable_compile():
    # The return value of torch._utils.is_compiling changes optimizer behavior.
    # We need that function to return True to include optimizer in the graph.
    # See: https://github.com/pytorch/pytorch/blob/a524123c91ab399c9dd6882c1189596dd77e7734/torch/optim/optimizer.py#L41
    def f_true():
        return True

    orig_is_compiling_code = torch._utils.is_compiling.__code__
    torch._utils.is_compiling.__code__ = f_true.__code__
    try:
        yield
    finally:
        torch._utils.is_compiling.__code__ = orig_is_compiling_code


def get_input_signature(*args, **kwargs):

    input_meta_str = pytree.tree_map(to_meta, [args, kwargs])
    return hashlib.sha256(repr(input_meta_str).encode("utf-8")).hexdigest()



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import functools
import operator

from easydist.metashard import metair

JAX_DEVICE_MESH = None

def set_device_mesh(device_mesh):
    global JAX_DEVICE_MESH
    JAX_DEVICE_MESH = device_mesh

    mesh_shape = device_mesh.device_ids.shape
    if mesh_shape[0] == 1:
        metair.DEVICE_MESH_1D = 0
    elif mesh_shape[1] == 1:
        metair.DEVICE_MESH_1D = 1


def get_device_mesh():
    global JAX_DEVICE_MESH
    return JAX_DEVICE_MESH


def device_mesh_world_size(device_mesh=None):
    if device_mesh is None:
        device_mesh = get_device_mesh()

    if device_mesh is None:
        return None

    device_mesh_shape = device_mesh.device_ids.shape

    return functools.reduce(operator.mul, device_mesh_shape)



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import copy
import logging

import numpy as np

from easydist import platform
from easydist.metashard.halo import HaloInfo, halo_padding
from easydist.metashard.annotation import ShardDim, ShardAnnotation
from easydist.metashard.combination import try_combination, HaloHint

logger = logging.getLogger(__name__)


def get_shard_size(dim_to_shard, shard_size):
    gcd = np.gcd.reduce(dim_to_shard)
    if gcd < shard_size:
        return gcd
    # elif gcd % shard_size == 0:
    #     return shard_size
    # return -1
    return shard_size


def get_shard_data(input_args, shard_size, dim_idx, haloinfo, chunk):
    chunk_data = platform.chunk(platform.clone(input_args), chunks=chunk, dim=dim_idx)
    chunk_shard_data = [
        platform.chunk(chunk_, chunks=shard_size, dim=dim_idx) for chunk_ in chunk_data
    ]
    shard_data = [
        platform.concatenate([t[i] for t in chunk_shard_data], dim_idx) for i in range(shard_size)
    ]

    return halo_padding(shard_data, haloinfo=haloinfo)


def check_prompt(flat_input_args, prompt_annotation):
    tensor_args = [i for i in flat_input_args if isinstance(flat_input_args, platform.Tensor)]
    if len(tensor_args) == len(prompt_annotation):
        for idx in range(len(tensor_args)):
            if tensor_args.ndim != len(prompt_annotation[idx]):
                return False
        return True
    return False


class MetaOp:

    def __init__(self, func, input_args, shard_size=2, name=None) -> None:
        self.func = func
        self.name = name
        if self.name is None:
            self.name = self.func.__name__
        self.input_args = input_args
        self.shard_size = shard_size
        self.flat_input_args, self.input_pytree_spec = platform.tree_flatten(input_args)

    def exec_platform(self, flat_input_args):
        if platform.get_backend() in ["torch", "tvm"]:
            args, kwargs = platform.tree_unflatten(flat_input_args, self.input_pytree_spec)
            return self.func(*args, **kwargs)
        elif platform.get_backend() == "jax":
            # (NOTE) disable_jit here for jax >= 4.7, the pjit primitive doesn't accept chunked input
            import jax
            with jax.disable_jit():
                flat_input_args = platform.tree_unflatten(flat_input_args, self.input_pytree_spec)
                if len(flat_input_args) == 3 and isinstance(flat_input_args[2], dict):
                    subfuns, invars, bind_params = flat_input_args
                    return self.func(*subfuns, *invars, **bind_params)
                return self.func(*flat_input_args)

    def exec(self, shard_annotation=None, priority_shard_dim_id=None, haloinfo=None):
        local_flat_input_args = self.flat_input_args

        tensor_list = [i for i in local_flat_input_args if isinstance(i, platform.Tensor)]

        if shard_annotation:
            dim_to_shard = []
            for tensor_ann_idx, tensor_ann in enumerate(shard_annotation):
                for dim_idx, dim in enumerate(tensor_ann):
                    if dim.shard_dim_id == priority_shard_dim_id:
                        dim_to_shard.append(tensor_list[tensor_ann_idx].shape[dim_idx])
                        break
            if len(dim_to_shard) == 0:
                raise RuntimeError(f"can not find priority_shard_dim_id {priority_shard_dim_id}")
            shard_size = get_shard_size(dim_to_shard, self.shard_size)
            if shard_size < self.shard_size:
                raise RuntimeError("can not find suitable shard_size")
            else:
                sharded_out = []
                for shard_idx in range(shard_size):
                    sharded_flat_input_args = []
                    tensor_arg = 0
                    for input_args in local_flat_input_args:
                        if isinstance(input_args, platform.Tensor):
                            tensor_ann = shard_annotation[tensor_arg]
                            no_shard_here = True
                            for dim_idx, dim in enumerate(tensor_ann):
                                if dim.shard_dim_id == priority_shard_dim_id:
                                    shard_data = get_shard_data(input_args, shard_size, dim_idx,
                                                                haloinfo, dim.chunk)[shard_idx]
                                    sharded_flat_input_args.append(shard_data)
                                    no_shard_here = False
                                    break
                            if no_shard_here:
                                sharded_flat_input_args.append(platform.clone(input_args))
                            tensor_arg += 1
                        else:
                            sharded_flat_input_args.append(input_args)

                    sharded_out.append(self.exec_platform(sharded_flat_input_args))
                return sharded_out

        else:
            return self.exec_platform(local_flat_input_args)

    def _try_sharding(self, fixed_annotation, subsequence_annotation, global_output, start_dim=0):
        shard_dim_id_flag = self.__shard_dim_id_flag

        if self.__find_new_strategy:
            return

        if len(subsequence_annotation) == 0:
            try:
                sharded_output = self.exec(shard_annotation=fixed_annotation,
                                           priority_shard_dim_id=shard_dim_id_flag)
            except RuntimeError as e:
                logger.debug(f"[{fixed_annotation}] {e}")
                return
            except:
                logger.debug(f"[{fixed_annotation}] run op.exec failed")
                return

            haloinfo = None
            combination_func = try_combination(sharded_output, global_output)
            if isinstance(combination_func, HaloHint):
                ecpt = combination_func
                if ecpt.halo <= 0:
                    ecpt.halo = 1
                if isinstance(sharded_output[0], tuple) or isinstance(sharded_output[0], list):
                    try_max_halo = sharded_output[0][ecpt.out_idx].shape[ecpt.dim] // 2
                else:
                    try_max_halo = sharded_output[0].shape[ecpt.dim] // 2
                for halo in range(ecpt.halo, try_max_halo):
                    sharded_output = self.exec(shard_annotation=fixed_annotation,
                                               priority_shard_dim_id=shard_dim_id_flag,
                                               haloinfo=HaloInfo(halo, ecpt.dim))
                    combination_func = try_combination(sharded_output, global_output)
                    if isinstance(combination_func, HaloHint):
                        combination_func = None
                    if combination_func:
                        haloinfo = HaloInfo(halo, ecpt.dim)
                        break

            if combination_func is not None and not isinstance(combination_func, HaloHint):
                self.__combination_ann[shard_dim_id_flag] = combination_func
                # inject haloinfo
                fixed_annotation.inject_haloinfo(haloinfo, shard_dim_id_flag)

                self.__sharding_annotion = copy.deepcopy(fixed_annotation)
                self.__find_new_strategy = True
            else:
                logger.debug(f"[{fixed_annotation}] combination failed")
            return

        for dim in range(start_dim, len(subsequence_annotation[0])):
            if subsequence_annotation[0][dim].shard_dim_id != 0:
                continue
            try_annotation = copy.deepcopy(subsequence_annotation[0])
            try_annotation[dim] = ShardDim.get_shard_dim(shard_dim_id_flag)
            self._try_sharding(fixed_annotation + ShardAnnotation([try_annotation]),
                               subsequence_annotation[1:], global_output)

        self._try_sharding(fixed_annotation + ShardAnnotation([subsequence_annotation[0]]),
                           subsequence_annotation[1:], global_output)

    def sharding_discovery(self, prompt_annotation=None):

        self.__combination_ann = {}
        # validate the prompt_annotation
        if prompt_annotation is not None and check_prompt(self.flat_input_args, prompt_annotation):
            max_shard_dim_id = prompt_annotation.get_max_shard_dim_id()
            for shard_dim_id in range(1, max_shard_dim_id + 1):
                combination_func = self.sharding_discovery_with_preset(
                    prompt_annotation, shard_dim_id)

                if combination_func is not None:
                    self.__combination_ann[shard_dim_id] = combination_func
                else:
                    break

        validate_max_shard_dim_id = len(self.__combination_ann)

        if validate_max_shard_dim_id >= 1:
            self.__sharding_annotion = prompt_annotation.clear_shard_dim(validate_max_shard_dim_id)
        else:
            # init the annotation with all NoShardDim
            self.__sharding_annotion = ShardAnnotation.init_from_input_args(self.flat_input_args)

        global_output = self.exec()

        self.__shard_dim_id_flag = validate_max_shard_dim_id + 1
        self.__find_new_strategy = False

        if len(self.__sharding_annotion) == 0:
            return self.__sharding_annotion, self.__combination_ann

        fixed_annotation = ShardAnnotation([])
        subsequence_annotation = self.__sharding_annotion
        start_dim = 0
        while True:
            self._try_sharding(fixed_annotation, subsequence_annotation, global_output, start_dim)
            if self.__find_new_strategy:

                # if shard_dim_id_flag not in sharding_annotion, then break
                max_shard_dim_id_now = 0
                for ann in self.__sharding_annotion:
                    for dim in ann:
                        max_shard_dim_id_now = max(max_shard_dim_id_now, dim.shard_dim_id)
                if max_shard_dim_id_now != self.__shard_dim_id_flag:
                    break

                self.__find_new_strategy = False
                # find the first dim for shard_dim_id_flag now
                # and update fixed_annotation, subsequence_annotation, start_dim
                fixed_annotation = ShardAnnotation([])
                finded = False
                for ann_idx, ann in enumerate(self.__sharding_annotion):
                    for idx, dim in enumerate(ann):
                        if self.__shard_dim_id_flag == dim.shard_dim_id:
                            finded = True
                            start_dim = idx + 1
                            break
                    if finded:
                        break
                    fixed_annotation += ShardAnnotation([ann])
                subsequence_annotation = self.__sharding_annotion[ann_idx:]
                if ann_idx == len(self.__sharding_annotion) - 1 and start_dim == len(
                        self.__sharding_annotion[-1]):
                    break
                self.__shard_dim_id_flag += 1
            else:
                break

        logger.debug(f"sharding_annotion of {self.name}: {self.__sharding_annotion}")

        return self.__sharding_annotion, self.__combination_ann

    def sharding_discovery_with_preset(self, sharding_annotion, priority_shard_dim_id=1):
        global_output = self.exec()

        try:
            sharded_output = self.exec(shard_annotation=sharding_annotion,
                                       priority_shard_dim_id=priority_shard_dim_id)
        except RuntimeError as e:
            logger.debug(f"[{sharding_annotion}] {e}")
            return
        except:
            logger.debug(f"[{sharding_annotion}] run op.exec failed")
            return

        combination_func = try_combination(sharded_output, global_output)
        if combination_func is not None and not isinstance(combination_func, HaloHint):
            return combination_func



# Most functions in this file are copied from https://github.com/pytorch/pytorch/blob/main/torch/distributed/_spmd/api.py.
# We copy these functions because they are not ready in released pytorch.

from functools import partial

import torch
from torch._decomp.decompositions import mse_loss, mse_loss_backward

aten = torch.ops.aten  # pyre-ignore

# from torch/distributed/_spmd/api.py
def _fused_adam_decomp(
    self,
    grads,
    exp_avgs,
    exp_avg_sqs,
    max_exp_avg_sqs,
    state_steps,
    *,
    lr=1,
    beta1=1,
    beta2=1,
    weight_decay=1,
    eps=1,
    amsgrad=True,
    maximize=True,
    grad_scale=None,
    found_inf=None,
):
    orig_tuple = (self, grads, exp_avgs, exp_avg_sqs, max_exp_avg_sqs)
    updated_tuple = aten._fused_adam.default(
        self,
        grads,
        exp_avgs,
        exp_avg_sqs,
        max_exp_avg_sqs,
        state_steps,
        lr=lr,
        beta1=beta1,
        beta2=beta2,
        weight_decay=weight_decay,
        eps=eps,
        amsgrad=amsgrad,
        maximize=maximize,
        grad_scale=grad_scale,
        found_inf=found_inf,
    )

    for idx, (orig, updated) in enumerate(zip(orig_tuple, updated_tuple)):
        if idx == 1:
            # skip gradient copying as we don't need to copy gradients back
            continue
        for o, u in zip(orig, updated):
            o.copy_(u)


# from torch/distributed/_spmd/api.py
def _foreach_add_decomp(self, other, alpha=1):
    self_updated = aten._foreach_add.List(self, other, alpha=alpha)
    for s, s_u in zip(self, self_updated):
        s.copy_(s_u)


# from torch/distributed/_spmd/api.py
def _foreach_unaop_decomp(op, self):
    self_updated = op(self)
    for s, s_u in zip(self, self_updated):
        s.copy_(s_u)


# from torch/distributed/_spmd/api.py
def _foreach_binop_list_decomp(op, self, other):
    self_updated = op(self, other)
    for s, s_u in zip(self, self_updated):
        s.copy_(s_u)


# from torch/distributed/_spmd/api.py
def _foreach_binop_scalar_decomp(op, self, scalar=1):
    self_updated = op(self, scalar)
    for s, s_u in zip(self, self_updated):
        s.copy_(s_u)


# from torch/distributed/_spmd/api.py
def _foreach_addcop_scalar_decomp(op, self, tensor1, tensor2, scalar=1):
    self_updated = op(self, tensor1, tensor2, scalar)
    for s, s_u in zip(self, self_updated):
        s.copy_(s_u)


# from torch/distributed/_spmd/api.py
def _foreach_addcop_tensor_decomp(op, self, tensor1, tensor2, tensor):
    self_updated = op(self, tensor1, tensor2, tensor)
    for s, s_u in zip(self, self_updated):
        s.copy_(s_u)


# modified from torch/distributed/_spmd/api.py
EASYDIST_DECOMP_TABLE = {
    aten._foreach_add_.List: _foreach_add_decomp,
    aten._foreach_add_.Scalar: partial(_foreach_binop_scalar_decomp, aten._foreach_add.Scalar),
    aten._foreach_addcdiv_.Scalar: partial(_foreach_addcop_scalar_decomp,
                                           aten._foreach_addcdiv.Scalar),
    aten._foreach_addcdiv_.Tensor: partial(_foreach_addcop_tensor_decomp,
                                           aten._foreach_addcdiv.Tensor),
    aten._foreach_addcmul_.Scalar: partial(_foreach_addcop_scalar_decomp,
                                           aten._foreach_addcmul.Scalar),
    aten._foreach_div_.List: partial(_foreach_binop_list_decomp, aten._foreach_div.List),
    aten._foreach_mul_.Scalar: partial(_foreach_binop_scalar_decomp, aten._foreach_mul.Scalar),
    aten._foreach_neg_.default: partial(_foreach_unaop_decomp, aten._foreach_neg.default),
    aten._foreach_reciprocal_.default: partial(_foreach_unaop_decomp,
                                               aten._foreach_reciprocal.default),
    aten._foreach_sub_.Scalar: partial(_foreach_binop_scalar_decomp, aten._foreach_sub.Scalar),
    aten._fused_adam_.default: _fused_adam_decomp,
    aten.mse_loss.default: mse_loss,
    aten.mse_loss_backward.default: mse_loss_backward,
}



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import logging
from functools import update_wrapper
from typing import Any

import numpy
import torch
import torch.utils._pytree as pytree
from torch.distributed._tensor import DeviceMesh

import easydist.config as mdconfig
from easydist.torch.device_mesh import (device_mesh_world_size, get_device_mesh, set_device_mesh)
from easydist.torch.experimental.compiler import _compile
from easydist.torch.experimental.init_helper import (CpuModuleInitHelper, SetParaInitHelper)
from easydist.torch.utils import get_input_signature

logger = logging.getLogger(__name__)


class CompiledFuncWrapper:

    def __init__(self,
                 func,
                 tracing_mode="fake",
                 cuda_graph=True,
                 enable_mono_graph=False) -> None:
        update_wrapper(self, func)
        self.original_func = func

        self.compiled_func = None
        self.tracing_mode = tracing_mode
        self.enable_cuda_graph = cuda_graph
        self.enable_mono_graph = enable_mono_graph

        self.init_helper = SetParaInitHelper()

        self.all_input_signature = []
        self.graph_list = {}

        self.cuda_graph_space = {}
        self.graph_pool = None

    def register_input_signature(self, *args: Any, **kwargs: Any) -> str:

        input_signature = get_input_signature(*args, **kwargs)

        if input_signature not in self.all_input_signature:

            self.all_input_signature.append(input_signature)
            logger.info(f"[Compile API] register input signature: {input_signature}")

            if self.enable_cuda_graph:
                self.cuda_graph_space[input_signature] = {
                    "cuda_graph": None,
                    "cuda_graph_input": None,
                    "cuda_graph_output": None,
                }

        return input_signature

    def register_cpu_module(self, cpu_module):
        self.init_helper = CpuModuleInitHelper(cpu_module)

    def set_init_helper(self, init_helper):
        self.init_helper = init_helper

    def __call__(self, *args: Any, **kwargs: Any) -> Any:

        input_signature = self.register_input_signature(*args, **kwargs)

        # skip compile when only one device
        world_size = torch.distributed.get_world_size()
        need_compile = world_size >= 2
        if get_device_mesh() is not None:
            need_compile = device_mesh_world_size() >= 2
        if need_compile and self.compiled_func is None:
            mesh_shape = numpy.array(range(world_size)).reshape(1, -1)
            mesh = DeviceMesh("cuda", mesh_shape.tolist())

            if get_device_mesh() == None:
                set_device_mesh(mesh)

            self.compiled_func = _compile(self.original_func, self.tracing_mode, self.init_helper,
                                          input_signature, args, kwargs)
            self.graph_list[input_signature] = self.compiled_func.graph
            # release the cpu module when finised pre-shard in _compiler
            self.init_helper = None

        def run_func(*args, **kwargs):
            if self.compiled_func:
                if input_signature not in self.graph_list:
                    if self.enable_mono_graph:
                        mono_graph = self.compiled_func.compile_mono_graph(*args, **kwargs)
                        self.graph_list[input_signature] = mono_graph
                        logger.info(f"[Compile API] compile mono graph for {input_signature}")
                    else:
                        raise RuntimeError(
                            "Input mismatch. If you are sure that different inputs do not change the graph, "
                            "you can try turning on the enable_mono_graph option.")
                graph = self.graph_list[input_signature]
                return self.compiled_func.run_with_graph(graph, *args, **kwargs)
            else:
                return self.original_func(*args, **kwargs)

        if self.enable_cuda_graph:
            current_space = self.cuda_graph_space[input_signature]
            if current_space["cuda_graph"] is None:
                logger.info("[Compile API] cuda graph warming up...")

                flatten_args, args_specs = pytree.tree_flatten([args, kwargs])
                current_space["cuda_graph_input"] = []
                for f_args in flatten_args:
                    if isinstance(f_args, torch.Tensor):
                        current_space["cuda_graph_input"].append(
                            torch.empty_like(f_args).copy_(f_args))
                    else:
                        current_space["cuda_graph_input"].append(f_args)
                args, kwargs = pytree.tree_unflatten(current_space["cuda_graph_input"], args_specs)

                # warm up
                s = torch.cuda.Stream()
                s.wait_stream(torch.cuda.current_stream())
                with torch.cuda.stream(s):
                    current_space["cuda_graph_output"] = run_func(*args, **kwargs)
                torch.cuda.current_stream().wait_stream(s)
                torch.cuda.synchronize()

                current_space["cuda_graph"] = torch.cuda.CUDAGraph()

                with torch.cuda.graph(current_space["cuda_graph"], self.graph_pool):
                    current_space["cuda_graph_output"] = run_func(*args, **kwargs)

                # assign the pool of first cuda graph as self.graph_pool
                if self.graph_pool is None:
                    self.graph_pool = current_space["cuda_graph"].pool()
            else:
                flatten_args, args_specs = pytree.tree_flatten([args, kwargs])
                for static_input, outside_input in zip(current_space["cuda_graph_input"],
                                                       flatten_args):
                    if isinstance(static_input, torch.Tensor):
                        static_input.copy_(outside_input)

            current_space["cuda_graph"].replay()
            return current_space["cuda_graph_output"]
        else:
            return run_func(*args, **kwargs)


def easydist_compile(func=None,
                     tracing_mode="fake",
                     cuda_graph=True,
                     enable_mono_graph=False,
                     use_hint=False,
                     liveness_only_input=False,
                     max_solver_time=float("inf")):

    mdconfig.use_hint = use_hint
    mdconfig.liveness_only_input = liveness_only_input
    mdconfig.max_seconds_same_incumbent = max_solver_time

    if func:
        return CompiledFuncWrapper(func, tracing_mode, cuda_graph, enable_mono_graph)
    else:

        def wrapper(func):
            return CompiledFuncWrapper(func, tracing_mode, cuda_graph, enable_mono_graph)

        return wrapper



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import List

from easydist import platform


class HaloInfo:

    def __init__(self, halowidth: int, dim: int) -> None:
        self.halowidth = halowidth
        self.dim = dim

    def __str__(self) -> str:
        return self.halowidth.__str__()

    def __repr__(self) -> str:
        return self.__str__()


def halo_padding(tensor_list_: List[platform.Tensor], haloinfo: HaloInfo) -> List[platform.Tensor]:
    """add halo padding to tensor_list_"""

    if haloinfo is None or len(tensor_list_) < 2:
        return tensor_list_

    halo = haloinfo.halowidth
    dim = haloinfo.dim

    padded_tensor_list = []
    for idx in range(len(tensor_list_)):
        to_concatenate = [tensor_list_[idx]]
        if idx >= 1:
            dim_size = tensor_list_[idx - 1].shape[dim]
            if dim_size < halo:
                raise RuntimeError("Cannot halo padding for this sharded_tensor")
            to_concatenate.insert(
                0, platform.narrow(tensor_list_[idx - 1], dim, dim_size - halo, halo))
        if idx <= len(tensor_list_) - 2:
            to_concatenate.append(platform.narrow(tensor_list_[idx + 1], dim, 0, halo))
        padded_tensor_list.append(platform.concatenate(to_concatenate, dim=dim))

    return padded_tensor_list



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from functools import partial

import torch
import torch.utils._pytree as pytree

add = torch.Tensor.add
equal = torch.equal
zeros_like = torch.zeros_like
min = torch.min
max = torch.max
allclose = partial(torch.allclose, rtol=1e-3, atol=1e-07)
concatenate = torch.concatenate
chunk = torch.chunk
narrow = torch.narrow

Tensor = torch.Tensor

tree_flatten = pytree.tree_flatten
tree_unflatten = pytree.tree_unflatten

clone = torch.clone
from_numpy = torch.from_numpy



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import os

import torch
from torch._functorch import config

from .api import compile, enable_transform, get_input_strategy, easydist_shard
from .bridge import shard_module, torch2meta_graph
from .passes.sharding import sharding_transform
from .device_mesh import set_device_mesh, get_device_mesh
from .sharding_interpreter import EDTorchShardingAnn
from .spmd_prop_rule import *

config.use_fake_tensor = False

__all__ = [
    'EDTorchShardingAnn', 'sharding_transform', 'set_device_mesh', 'get_device_mesh',
    'torch2meta_graph', 'compile', 'easydist_shard', 'enable_transform', 'shard_module',
    'get_input_strategy'
]


def easydist_setup_torch(device, allow_tf32):
    torch.backends.cuda.matmul.allow_tf32 = allow_tf32
    torch.backends.cudnn.allow_tf32 = allow_tf32

    # this environment used for cuda graph
    os.environ["NCCL_ASYNC_ERROR_HANDLING"] = "0"



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import functools
from enum import Enum
import logging

from easydist import platform
import easydist.config as mdconfig

logger = logging.getLogger(__name__)

TRY_COMBINATION_FUNC = dict()


def try_combination_func(func):
    """Register a function as a try_combination_func"""
    TRY_COMBINATION_FUNC[func.__name__] = func
    return func


class ReduceOp(Enum):
    SUM = 1
    MAX = 2
    MIN = 3


class HaloHint:

    def __init__(self, halo, dim, idx_=None):
        self.halo = halo
        self.dim = dim
        self.out_idx = idx_


def aligned_prefix(t1: platform.Tensor, t2: platform.Tensor, dim_idx: int):
    """calculate the max prefix for t1 and t2 on dimension(dim_idx)

    For example: aligned_prefix([1, 2, 3], [1, 2, 5]) -> 2
    """
    dim_size = min(t1.shape[dim_idx], t2.shape[dim_idx])
    for i in range(1, dim_size + 1):
        if not platform.allclose(platform.narrow(t1, dim_idx, 0, i),
                                 platform.narrow(t2, dim_idx, 0, i)):
            return i - 1
    return i


def shape_aligned_otherdim(shape_1, shape_2, dim_idx):
    """return True when shape_1 and shape_2 only different on dimension(dim_idx)"""
    if len(shape_1) != len(shape_2):
        return False

    diff_dim = []
    for idx in range(len(shape_1)):
        if shape_1[idx] != shape_2[idx]:
            diff_dim.append(idx)

    if diff_dim == [dim_idx]:
        return True
    return False


class CombinationFunc:

    @staticmethod
    def identity(sharded_tensor):
        identity_tensor = sharded_tensor[0]
        for tensor_ in sharded_tensor:
            if not platform.equal(identity_tensor, tensor_):
                logger.debug("not all tensor same as identity")
                return None
        return identity_tensor

    @staticmethod
    def reduce(sharded_tensor, ops=ReduceOp.SUM):
        init = platform.zeros_like(sharded_tensor[0])
        assert ops in [ReduceOp.SUM, ReduceOp.MAX, ReduceOp.MIN]
        if ops == ReduceOp.SUM:
            reduce_func_ = platform.add
        elif ops == ReduceOp.MAX:
            reduce_func_ = platform.max
        elif ops == ReduceOp.MIN:
            reduce_func_ = platform.min
        return functools.reduce(reduce_func_, sharded_tensor, init)

    @staticmethod
    def gather(sharded_tensor, dim, halowidth=0, chunk=1):
        if halowidth == 0:
            if chunk == 1:
                return platform.concatenate(sharded_tensor, dim=dim)
            else:
                chunk_sharded_tensor = [platform.chunk(t, chunk, dim) for t in sharded_tensor]
                reorder_list = []
                for chunk_idx in range(chunk):
                    reorder_list += [chunk_list[chunk_idx] for chunk_list in chunk_sharded_tensor]
                return platform.concatenate(reorder_list, dim=dim)

        shard_world_size = len(sharded_tensor)
        gathered_tensor = sharded_tensor[0]
        for idx_ in range(1, shard_world_size):
            width_tensor_1 = int(gathered_tensor.shape[dim])
            width_tensor_2 = int(sharded_tensor[idx_].shape[dim])
            if halowidth > 0:
                gathered_tensor = platform.concatenate([
                    platform.narrow(gathered_tensor, dim, 0, width_tensor_1 - halowidth),
                    platform.add(
                        platform.narrow(gathered_tensor, dim, width_tensor_1 - halowidth,
                                        halowidth),
                        platform.narrow(sharded_tensor[idx_], dim, 0, halowidth)),
                    platform.narrow(sharded_tensor[idx_], dim, halowidth,
                                    width_tensor_2 - halowidth)
                ],
                                                       dim=dim)
            else:
                gathered_tensor = platform.concatenate([
                    platform.narrow(gathered_tensor, dim, 0, width_tensor_1 + halowidth),
                    platform.narrow(sharded_tensor[idx_], dim, -1 * halowidth,
                                    width_tensor_2 + halowidth)
                ],
                                                       dim=dim)

        return gathered_tensor


@try_combination_func
def try_combination_identity(sharded_tensor_, global_tensor_):
    """try combination through identity, return combination func if success, otherwise None"""

    for t_ in sharded_tensor_:
        if t_.shape != global_tensor_.shape:
            return None

    result = CombinationFunc.identity(sharded_tensor_)
    if result is not None and platform.allclose(result, global_tensor_):
        return functools.partial(CombinationFunc.identity)

    return None


@try_combination_func
def try_combination_reduce(sharded_tensor_, global_tensor_):
    """try combination through reduce, return combination func if success, otherwise None"""

    for t_ in sharded_tensor_:
        if t_.shape != global_tensor_.shape:
            return None

    for reduce_op in [ReduceOp.SUM, ReduceOp.MAX, ReduceOp.MIN]:
        reduce_func = functools.partial(CombinationFunc.reduce, ops=reduce_op)
        if platform.allclose(reduce_func(sharded_tensor_), global_tensor_):
            return reduce_func

    return None


@try_combination_func
def try_combination_gather(sharded_output_, global_output_):
    """try combination through gather, return combination func if success, otherwise None"""

    # 1) all sharded tensor shape should be equal to global tensor except the gather_dim
    sharded_output_shape = sharded_output_[0].shape
    global_output_shape = global_output_.shape

    # return if global_output is only one element.
    if len(global_output_shape) == 0:
        return None

    shard_size = len(sharded_output_)

    for dim_idx in range(len(sharded_output_shape)):
        if sharded_output_shape[dim_idx] != global_output_shape[dim_idx]:
            break

    for tensor_ in sharded_output_:
        if not shape_aligned_otherdim(tensor_.shape, global_output_shape, dim_idx):
            return None

    combination_size = 0
    for out_ in sharded_output_:
        combination_size += out_.shape[dim_idx]
    gap_size = combination_size - global_output_shape[dim_idx]
    gather_func = None
    if gap_size == 0:
        gather_func = functools.partial(CombinationFunc.gather, dim=dim_idx)
        gathered_tensor = gather_func(sharded_output_)
        if gathered_tensor.shape == global_output_.shape:
            if platform.allclose(gathered_tensor, global_output_):
                return gather_func
            if mdconfig.extend_space:
                reference_shard_output = platform.chunk(global_output_,
                                                        chunks=shard_size,
                                                        dim=dim_idx)[0]
                aligned_prefix_len = aligned_prefix(sharded_output_[0], reference_shard_output,
                                                    dim_idx)

                # explore gather with chunk > 1
                if aligned_prefix_len != 0 and sharded_output_shape[
                        dim_idx] % aligned_prefix_len == 0:
                    guess_chunk = sharded_output_shape[dim_idx] // aligned_prefix_len
                    gather_func = functools.partial(CombinationFunc.gather,
                                                    dim=dim_idx,
                                                    chunk=guess_chunk)
                    gathered_tensor_chunk = gather_func(sharded_output_)
                    if platform.allclose(gathered_tensor_chunk, global_output_):
                        return gather_func

                # return HaloHint to explore the chance of halo sharding
                if aligned_prefix_len > (sharded_output_shape[dim_idx] // 2):
                    return HaloHint(sharded_output_shape[dim_idx] - aligned_prefix_len, dim_idx)

    if mdconfig.extend_space:
        # halo gather with halowidth > 0
        if gap_size > 0 and gap_size % (shard_size - 1) == 0:
            halowidth = int(gap_size // (shard_size - 1))
            if halowidth >= combination_size // shard_size:
                return None
            gather_func = functools.partial(CombinationFunc.gather,
                                            dim=dim_idx,
                                            halowidth=halowidth)
            gathered_tensor = gather_func(sharded_output_)
            if gathered_tensor.shape == global_output_.shape:
                if platform.allclose(gathered_tensor, global_output_):
                    return gather_func

        # halo gather with halowidth < 0
        if gap_size > 0 and gap_size % shard_size == 0:
            halowidth = -1 * int(gap_size // shard_size)
            if -1 * halowidth >= combination_size // (2 * shard_size):
                return None
            gather_func = functools.partial(CombinationFunc.gather,
                                            dim=dim_idx,
                                            halowidth=halowidth)
            gathered_tensor = gather_func(sharded_output_)
            if gathered_tensor.shape == global_output_.shape:
                if platform.allclose(gathered_tensor, global_output_):
                    return gather_func

        # raise HaloHint when convolution without padding
        if gap_size < 0 and gap_size % (shard_size - 1) == 0:
            halowidth = int(gap_size // (shard_size - 1)) // 2
            if -1 * halowidth >= combination_size // shard_size:
                return None
            return HaloHint(halowidth, dim_idx)


def try_combination_single(sharded_output_, global_output_):

    # check all sharded tensor have equal dimension of global_output
    for sharded_tensor in sharded_output_:
        if len(sharded_tensor.shape) != len(global_output_.shape):
            return None

    for func_name in TRY_COMBINATION_FUNC:
        combination_func = TRY_COMBINATION_FUNC[func_name](sharded_output_, global_output_)
        if combination_func:
            return combination_func

    return None


def try_combination(sharded_output_, global_output_):

    if isinstance(global_output_, platform.Tensor):
        return try_combination_single(sharded_output_, global_output_)

    if isinstance(global_output_, tuple) or isinstance(global_output_, list):
        output_num = [len(i) for i in sharded_output_]
        if len(global_output_) == min(output_num) == max(output_num):
            return_combination_ann = []
            for idx_ in range(len(global_output_)):
                if isinstance(global_output_[idx_], platform.Tensor):
                    single_ann = try_combination_single([i[idx_] for i in sharded_output_],
                                                        global_output_[idx_])
                    if single_ann is None:
                        return None
                    if isinstance(single_ann, HaloHint):
                        single_ann.out_idx = idx_
                        return single_ann
                    return_combination_ann.append(single_ann)
                else:
                    for out_ in sharded_output_:
                        if global_output_[idx_] != out_[idx_]:
                            return None

            if len(return_combination_ann) > 0:
                return return_combination_ann

    return None



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import functools

import torch
from torch.fx.node import _get_qualified_name

from easydist.metashard import view_propagation
from easydist.metashard.annotation import ShardAnnotation, ShardDim
from easydist.metashard.combination import CombinationFunc, ReduceOp

from .device_mesh import device_mesh_world_size

aten = torch.ops.aten


class PresetMetaSPMD:
    op_rules = {}


def register_meta_spmd(op):

    def wrapper(impl):
        op_list = op if isinstance(op, list) else [op]
        for op_i in op_list:
            if isinstance(op_i, str):
                op_name = op_i
            else:
                op_name = _get_qualified_name(op_i)
            PresetMetaSPMD.op_rules[op_name] = impl
        return impl

    return wrapper


def preset_meta_spmd(meta_op, input_args=None):
    if isinstance(meta_op, str):
        args, kwargs = input_args
        return PresetMetaSPMD.op_rules[meta_op](args, kwargs)
    elif meta_op.name in PresetMetaSPMD.op_rules:
        if input_args is None:
            args, kwargs = meta_op.input_args
        else:
            args, kwargs = input_args
        return PresetMetaSPMD.op_rules[meta_op.name](args, kwargs)
    return None, None


@register_meta_spmd([
    "placeholder",
    "get_attr",
])
def meta_spmd_placeholder(args, kwargs):
    input_shape = args.shape
    output_shape = args.shape

    view_ann = view_propagation(input_shape, output_shape, world_size=device_mesh_world_size())
    return view_ann['sharding_ann'], view_ann['combination_ann']


@register_meta_spmd([
    aten.view.default,
    aten._unsafe_view.default,
])
def meta_spmd_view(args, kwargs):
    input_shape = args[0].shape
    output_shape = list(args[1])

    view_ann = view_propagation(input_shape, output_shape, world_size=device_mesh_world_size())
    return view_ann['sharding_ann'], view_ann['combination_ann']


@register_meta_spmd(aten._reshape_alias.default)
def meta_spmd_reshape(args, kwargs):
    input_shape = args[0].shape
    output_shape = list(args[1])
    stride = list(args[2])

    view_ann = view_propagation(input_shape, output_shape, world_size=device_mesh_world_size())
    return view_ann['sharding_ann'], view_ann['combination_ann']


# (TODO) here just replicate all optimizer computation
# need to generate some shard strategy here
@register_meta_spmd([
    aten._foreach_add.Scalar,
    aten._foreach_mul.Scalar,
    aten._foreach_sub.Scalar,
    aten._foreach_div.Scalar,
    aten._foreach_neg.default,
    aten._foreach_reciprocal.default,
    aten._foreach_sqrt.default,
])
def meta_foreach_one_tensor_list(args, kwargs):
    tensor_list = args[0]
    sharding_ann = ShardAnnotation([[[ShardDim.get_noshard_dim()] * len(tensor.shape)
                                     for tensor in tensor_list]])
    return sharding_ann, {}


@register_meta_spmd([
    aten._foreach_add.List,
    aten._foreach_mul.List,
    aten._foreach_sub.List,
    aten._foreach_div.List,
])
def meta_foreach_two_tensor_list(args, kwargs):
    tensor_1_list, tensor_2_list = args[0], args[1]
    sharding_ann = ShardAnnotation([[[ShardDim.get_noshard_dim()] * len(tensor.shape)
                                     for tensor in tensor_1_list],
                                    [[ShardDim.get_noshard_dim()] * len(tensor.shape)
                                     for tensor in tensor_2_list]])
    return sharding_ann, {}


@register_meta_spmd([
    aten._foreach_addcdiv.Scalar,
    aten._foreach_addcmul.Scalar,
])
def meta_foreach_three_tensor_list(args, kwargs):
    tensor_1_list, tensor_2_list, tensor_3_list = args[0], args[1], args[2]
    sharding_ann = ShardAnnotation([[[ShardDim.get_noshard_dim()] * len(tensor.shape)
                                     for tensor in tensor_1_list],
                                    [[ShardDim.get_noshard_dim()] * len(tensor.shape)
                                     for tensor in tensor_2_list],
                                    [[ShardDim.get_noshard_dim()] * len(tensor.shape)
                                     for tensor in tensor_3_list]])
    return sharding_ann, {}


@register_meta_spmd([aten.empty.memory_format, aten.zeros.default])
def meta_create_op(args, kwargs):
    tensor_shape = args[0]

    shard_idx = 1
    combine_ann = {}

    sharding_ann = ShardAnnotation([[ShardDim.get_noshard_dim() for _ in tensor_shape]])

    world_size = device_mesh_world_size()

    for dim_idx, dim_shape in enumerate(tensor_shape):
        if world_size <= dim_shape:
            sharding_ann[0][dim_idx] = ShardDim(shard_idx)
            combine_ann[shard_idx] = functools.partial(CombinationFunc.gather, dim=dim_idx)
            shard_idx += 1

    return sharding_ann, combine_ann


@register_meta_spmd(aten.cat.default)
def meta_cat(args, kwargs):
    tensorlist, dim = args[0:2]
    tensor_shape = tensorlist[0].shape

    shard_idx = 1
    sharding_ann = [ShardDim.get_noshard_dim() for _ in tensor_shape]
    combine_ann = {}

    world_size = device_mesh_world_size()

    for idx in range(len(tensor_shape)):
        if world_size <= tensor_shape[idx] and idx != dim:
            sharding_ann[idx] = ShardDim(shard_idx)
            combine_ann[shard_idx] = functools.partial(CombinationFunc.gather, dim=idx)
            shard_idx += 1

    sharding_ann = ShardAnnotation([sharding_ann for _ in tensorlist])

    return sharding_ann, combine_ann


@register_meta_spmd(aten._log_softmax_backward_data.default)
def meta_log_softmax_backward_data(args, kwargs):
    grad_output, output, dim = args[0:3]
    tensor_shape = grad_output.shape

    shard_idx = 1
    sharding_ann = [ShardDim.get_noshard_dim() for _ in tensor_shape]
    combine_ann = {}

    world_size = device_mesh_world_size()

    for idx in range(len(tensor_shape)):
        if world_size <= tensor_shape[idx] and idx != dim:
            sharding_ann[idx] = ShardDim(shard_idx)
            combine_ann[shard_idx] = functools.partial(CombinationFunc.gather, dim=idx)
            shard_idx += 1

    sharding_ann = ShardAnnotation([sharding_ann] * 2)

    return sharding_ann, combine_ann


@register_meta_spmd(aten.mm.default)
def meta_mm(args, kwargs):
    mat1, mat2 = args[0:2]
    shape_n, shape_m = mat1.shape
    shape_m, shape_p = mat2.shape

    sharding_ann = ShardAnnotation([[ShardDim.get_noshard_dim() for _ in range(2)]
                                    for _ in range(2)])
    shard_idx = 1
    combine_ann = {}

    world_size = device_mesh_world_size()

    if world_size <= shape_n:
        sharding_ann[0][0] = ShardDim(shard_idx)
        combine_ann[shard_idx] = functools.partial(CombinationFunc.gather, dim=0)
        shard_idx += 1

    if world_size <= shape_m:
        sharding_ann[0][1] = ShardDim(shard_idx)
        sharding_ann[1][0] = ShardDim(shard_idx)
        combine_ann[shard_idx] = functools.partial(CombinationFunc.reduce, ops=ReduceOp.SUM)
        shard_idx += 1

    if world_size <= shape_p:
        sharding_ann[1][1] = ShardDim(shard_idx)
        combine_ann[shard_idx] = functools.partial(CombinationFunc.gather, dim=1)
        shard_idx += 1

    return sharding_ann, combine_ann


@register_meta_spmd(aten.bmm.default)
def meta_bmm(args, kwargs):
    mat1, mat2 = args[0:2]
    shape_b_, shape_n, shape_m = mat1.shape
    shape_b, shape_m, shape_p = mat2.shape
    assert shape_b_ == shape_b

    sharding_ann = ShardAnnotation([[ShardDim.get_noshard_dim() for _ in range(3)]
                                    for _ in range(2)])
    shard_idx = 1
    combine_ann = {}

    world_size = device_mesh_world_size()

    if world_size <= shape_b:
        sharding_ann[0][0] = ShardDim(shard_idx)
        sharding_ann[1][0] = ShardDim(shard_idx)
        combine_ann[shard_idx] = functools.partial(CombinationFunc.gather, dim=0)
        shard_idx += 1

    if world_size <= shape_n:
        sharding_ann[0][1] = ShardDim(shard_idx)
        combine_ann[shard_idx] = functools.partial(CombinationFunc.gather, dim=1)
        shard_idx += 1

    if world_size <= shape_m:
        sharding_ann[0][2] = ShardDim(shard_idx)
        sharding_ann[1][1] = ShardDim(shard_idx)
        combine_ann[shard_idx] = functools.partial(CombinationFunc.reduce, ops=ReduceOp.SUM)
        shard_idx += 1

    if world_size <= shape_p:
        sharding_ann[1][2] = ShardDim(shard_idx)
        combine_ann[shard_idx] = functools.partial(CombinationFunc.gather, dim=2)
        shard_idx += 1

    return sharding_ann, combine_ann


@register_meta_spmd(aten.empty_like.default)
def meta_empty_like(args, kwargs):
    tensor_shape = args[0].shape

    world_size = device_mesh_world_size()

    sharding_ann = ShardAnnotation([[ShardDim.get_noshard_dim()] * len(tensor_shape)])
    shard_idx = 1
    combine_ann = {}
    for idx, shape in enumerate(tensor_shape):
        if world_size <= shape:
            sharding_ann[0][idx] = ShardDim(shard_idx)
            combine_ann[shard_idx] = functools.partial(CombinationFunc.gather, dim=idx)
            shard_idx += 1

    return sharding_ann, combine_ann


@register_meta_spmd(aten.native_group_norm.default)
def meta_group_norm(args, kwargs):
    tensor_shape, w_shape, b_shape = args[0].shape, args[1].shape, args[2].shape

    world_size = device_mesh_world_size()
    sharding_ann = ShardAnnotation([[ShardDim.get_noshard_dim()] * len(tensor_shape),
                                    [ShardDim.get_noshard_dim()] * len(w_shape),
                                    [ShardDim.get_noshard_dim()] * len(b_shape)])

    shard_idx = 1
    combine_ann = {}

    if world_size <= tensor_shape[0]:
        sharding_ann[0][0] = ShardDim(shard_idx)
        combine_ann[shard_idx] = [
            functools.partial(CombinationFunc.gather, dim=0),
            functools.partial(CombinationFunc.gather, dim=0),
            functools.partial(CombinationFunc.gather, dim=0)
        ]

    return sharding_ann, combine_ann



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from .fix_embedding import fix_embedding
from .fix_bias import fix_addmm_bias, fix_convoluation_bias
from .fix_meta_device import fix_meta_device
from .eliminate_detach import eliminate_detach
from .sharding import sharding_transform

__all__ = [
    "fix_embedding", "fix_addmm_bias", "fix_convoluation_bias", "eliminate_detach",
    "sharding_transform", "fix_meta_device"
]



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import os
import time

from easydist.platform import get_backend
import easydist.config as mdconfig


class EDTimer:

    def __init__(self, func, trials=3, warmup_trials=3, in_ms=True, device=None) -> None:
        self.func = func
        self.warmup_trials = warmup_trials
        self.trials = trials
        self.in_ms = in_ms

        self.device = device
        if self.device == None:
            self.device = mdconfig.easydist_device

        self.backend = get_backend()

    def time(self):
        if self.backend == "jax":
            return self.time_jax()
        elif self.backend == "torch":
            if self.device == "cuda":
                return self.time_torch_cuda()
            elif self.device == "cpu":
                return self.time_cpu()
        return None

    def time_cpu(self):
        for _ in range(self.warmup_trials):
            self.func()

        start_t = time.perf_counter()
        for _ in range(self.trials):
            self.func()

        elapsed_time = time.perf_counter() - start_t
        elapsed_time = elapsed_time / self.trials

        # time elapsed in **milliseconds**
        if self.in_ms:
            return elapsed_time * 1000
        return elapsed_time

    def time_torch_cuda(self):
        import torch

        start_evt = []
        end_evt = []
        for _ in range(0, self.trials):
            start_evt.append(torch.cuda.Event(enable_timing=True))
            end_evt.append(torch.cuda.Event(enable_timing=True))

        for trial_idx in range(0, self.trials + self.warmup_trials):
            evt_idx = trial_idx - self.warmup_trials

            if evt_idx >= 0:
                start_evt[evt_idx].record()

            self.func()

            if evt_idx >= 0:
                end_evt[evt_idx].record()

        torch.cuda.synchronize()
        ops_elapsed_time = 0
        for evt_idx in range(0, self.trials):
            # time elapsed in **milliseconds**
            ops_elapsed_time += start_evt[evt_idx].elapsed_time(end_evt[evt_idx])
        ops_elapsed_time = ops_elapsed_time / self.trials

        if self.in_ms:
            return ops_elapsed_time
        return ops_elapsed_time / 1000

    def time_jax(self):
        import jax
        for _ in range(self.warmup_trials):
            self.func()
            (jax.device_put(0.) + 0).block_until_ready()

        start_t = time.perf_counter()
        for _ in range(self.trials):
            self.func()
            (jax.device_put(0.) + 0).block_until_ready()

        elapsed_time = time.perf_counter() - start_t
        elapsed_time = elapsed_time / self.trials

        # time elapsed in **milliseconds**
        if self.in_ms:
            return elapsed_time * 1000
        return elapsed_time



# Copyright (c) 2023, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import functools
import logging
import time
from typing import Any

import jax
import rich
from jax import core
from jax._src.distributed import global_state
from jax._src.util import safe_map
from jax.experimental import mesh_utils, multihost_utils
from jax.sharding import Mesh, NamedSharding, PartitionSpec

import easydist.config as mdconfig
from easydist.autoflow import AutoFlowSolver
from easydist.metashard import metair

from .device_mesh import set_device_mesh, get_device_mesh
from .bridge import jax2md_bridge
from .sharding_interpreter import EDJaxShardingAnn
from .utils import _sharding_ann_env

logger = logging.getLogger(__name__)


INPUT_STRATEGY = None


def to_shape_array(x):
    if isinstance(x, jax.Array) and not jax.core.is_opaque_dtype(x.dtype):
        return core.ShapedArray(shape=x.shape, dtype=x.dtype)
    else:
        return x


def materialize(x):
    if isinstance(x, core.ShapedArray):
        key = jax.random.PRNGKey(seed=42)
        if x.dtype.name in ["float64", "float32", "float16"]:
            return jax.random.normal(key, shape=x.shape, dtype=x.dtype)
        elif x.dtype.name in ["int32", "unint32", "int64", "uint64", "uint8"]:
            return jax.random.randint(key, shape=x.shape, dtype=x.dtype, minval=1, maxval=8)
        elif x.dtype.name in ["bool"]:
            return jax.random.normal(key, shape=x.shape) > 1.
        else:
            return jax.numpy.empty(shape=x.shape, dtype=x.dtype)
    return x


def convert(strategy, mesh, val):
    axis_names = mesh.axis_names

    s1, s2 = strategy[0], strategy[1]
    ndim = len(val.shape)
    mesh_shape = [None] * ndim

    # we use strategy except val.shape -> ()
    if ndim > 0:
        for idx, s_ in enumerate([s1, s2]):
            if s_.state == metair.SPMD.SHARD:
                dim = s_.args["dim"]
                if mesh_shape[dim] == None:
                    mesh_shape[dim] = axis_names[idx]
                else:
                    mesh_shape[dim] = mesh.axis_names

    return NamedSharding(mesh, PartitionSpec(*mesh_shape))


def _get_local_array(array, specs):
    mesh = get_device_mesh()
    mesh_shape = mesh.device_ids.shape
    axis_names = mesh.axis_names
    for idx in range(len(specs)):
        for axis_idx, axis in enumerate(axis_names):
            if specs[idx] == axis:
                global_rank = global_state.process_id
                rank_coords = (mesh.device_ids == global_rank).nonzero()
                mesh_dim_rank = rank_coords[axis_idx].item()
                array = jax.numpy.array_split(array, mesh_shape[axis_idx], axis=idx)[mesh_dim_rank]

    return array


def shard_module(flatten_args):

    device_mesh = get_device_mesh()

    for i in range(len(flatten_args)):
        if isinstance(flatten_args[i], jax.Array) or isinstance(flatten_args[i], jax.ShapedArray):
            strategy = convert(INPUT_STRATEGY[i], device_mesh, flatten_args[i])
            flatten_args[i] = materialize(flatten_args[i])
            local_array = _get_local_array(flatten_args[i], strategy.spec)
            flatten_args[i] = multihost_utils.host_local_array_to_global_array(
                local_array, device_mesh, strategy.spec)

    return flatten_args


def add_sharding_jaxpr(jaxpr, consts, shard_strategy, args):
    env = {}

    def read(var):
        if type(var) is core.Literal:
            return var.val
        return env[var]

    def write(var, val):
        env[var] = val

    # Args now correspond to Jaxpr outvars
    safe_map(write, jaxpr.invars, args)
    safe_map(write, jaxpr.constvars, consts)

    # Looping backward
    for eqn in jaxpr.eqns:
        subfuns, bind_params = eqn.primitive.get_bind_params(eqn.params)

        onelevel_key = eqn.primitive.__str__()
        if eqn.primitive.__str__() == "custom_jvp_call":
            onelevel_key += "[" + \
                subfuns[0].f.args[0].eqns[0].params['name'] + "]"
        if eqn.primitive.__str__() == "xla_call":
            onelevel_key += "[" + eqn.params['name'] + "]"

        invars = [var for var in eqn.invars if isinstance(var, jax.core.Var)]

        #  outvars are now invars
        invals = safe_map(read, eqn.invars)

        # NOTE, use last outvars for name of MetaNode in Jax
        unique_key = str(eqn.outvars[-1])

        mesh = get_device_mesh()

        if unique_key in shard_strategy:
            in_strtg_group = shard_strategy[unique_key]['strategy'].in_strtg_group

            sharding_idx = 0
            for idx in range(len(invals)):
                if isinstance(invals[idx], jax.interpreters.partial_eval.DynamicJaxprTracer):
                    strategy_ = convert(in_strtg_group.get_var_strtg(sharding_idx), mesh,
                                        invals[idx])
                    invals[idx] = jax.lax.with_sharding_constraint(invals[idx], strategy_)
                    sharding_idx += 1

        outval = eqn.primitive.bind(*subfuns, *invals, **bind_params)

        if isinstance(outval, jax.interpreters.partial_eval.DynamicJaxprTracer):
            outval = [outval]

        safe_map(write, eqn.outvars, outval)

    outvals = safe_map(read, jaxpr.outvars)

    return outvals


def get_opt_strategy(func, *args, **kwargs):

    global INPUT_STRATEGY

    closed_jaxpr = jax.make_jaxpr(func)(*args, **kwargs)

    # use tf32 during MetaSPMD Annotaion
    with _sharding_ann_env():

        start_t = time.perf_counter()
        sharding_interpreter = EDJaxShardingAnn(closed_jaxpr.jaxpr)
        sharding_info, shape_info = sharding_interpreter.run(closed_jaxpr.literals, *args,
                                                             **kwargs)
        logger.info(f"[EDJaxShardingAnn.run]: {time.perf_counter() - start_t} s.")

        if mdconfig.log_level <= logging.DEBUG and global_state.process_id == 0:
            rich.print("sharding_info:\n", sharding_info)
            rich.print("shape_info:\n", shape_info)

    meta_graph = jax2md_bridge(closed_jaxpr.jaxpr, sharding_info, shape_info)

    if mdconfig.log_level <= logging.DEBUG and global_state.process_id == 0:
        rich.print(meta_graph)

    device_mesh = get_device_mesh()
    solver = AutoFlowSolver(device_mesh.device_ids.shape)
    solver.add_graph(meta_graph)
    start_t = time.perf_counter()
    opt_strategy = solver.ilp_optimize()
    logger.info(f"[AutoFlowSolver.ilp_optimize]: {time.perf_counter() - start_t} s.")
    # start_t = time.perf_counter()
    # beam_search_strategy = solver.beam_search()
    # logger.info(f"[AutoFlowSolver.beam_search]: {time.perf_counter() - start_t} s.")

    INPUT_STRATEGY = meta_graph.get_input_strategy(opt_strategy)

    if mdconfig.log_level <= logging.DEBUG and global_state.process_id == 0:
        rich.print(opt_strategy)

    return opt_strategy


def easydist_shard(fun, shard_strategy={}):

    @functools.wraps(fun)
    def wrapped(*args, **kwargs):
        # Since we assume unary functions, we won't worry about flattening and
        # unflattening arguments.
        closed_jaxpr, return_shape = jax.make_jaxpr(fun, return_shape=True)(*args, **kwargs)

        flatten_args, in_tree = jax.tree_util.tree_flatten(args)

        out = add_sharding_jaxpr(closed_jaxpr.jaxpr, closed_jaxpr.literals, shard_strategy,
                                 flatten_args)

        flatten_shape, shape_tree = jax.tree_util.tree_flatten(return_shape)

        out_idx = 0
        for idx, shape in enumerate(flatten_shape):
            if isinstance(shape, jax.ShapeDtypeStruct):
                flatten_shape[idx] = out[out_idx]
                out_idx += 1

        out_tree = jax.tree_util.tree_unflatten(shape_tree, flatten_shape)

        return out_tree

    return wrapped


# =============== Experimetal API for Jax ==================


def _compile(func, args, kwargs):

    # setup the device mesh
    size = jax.device_count()
    devices = mesh_utils.create_device_mesh((1, size))
    mesh = Mesh(devices, axis_names=('a', 'b'))

    set_device_mesh(mesh)

    opt_strategy = get_opt_strategy(func, *args, **kwargs)

    shard_func = jax.jit(easydist_shard(func, opt_strategy))

    if mdconfig.log_level <= logging.DEBUG and global_state.process_id == 0:
        closed_jaxpr = jax.make_jaxpr(shard_func)(*args, **kwargs)
        print(closed_jaxpr.jaxpr)
        print(closed_jaxpr.literals)

    def compiled_func_return(*args, **kwargs):

        flatten_args, specs = jax.tree_util.tree_flatten((args, kwargs))
        flatten_args = shard_module(flatten_args)
        shard_args, shard_kwargs = jax.tree_util.tree_unflatten(specs, flatten_args)

        with jax.spmd_mode('allow_all'):
            out = shard_func(*shard_args, **shard_kwargs)

        return out

    return compiled_func_return


class CompiledFuncWrapper:

    def __init__(self, func) -> None:
        self.original_func = func
        self.compiled_func = None

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        if self.compiled_func is None:

            self.compiled_func = _compile(self.original_func, args, kwargs)

        return self.compiled_func(*args, **kwargs)


def easydist_compile(func=None, max_solver_time=float("inf"), liveness_only_input=False):

    mdconfig.liveness_only_input = liveness_only_input
    mdconfig.max_seconds_same_incumbent = max_solver_time

    if func:
        return CompiledFuncWrapper(func)

    else:

        def wrapper(func):
            return CompiledFuncWrapper(func)

        return wrapper
